# Balazar
# Copyright (C) 2003-2004 Jean-Baptiste LAMY
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import sys, math
import soya

from balazar.character       import *
from balazar.item            import Item, EquipableItem, UseableItem, Food, Camera, Photo
from balazar.vegetation      import Vegetal, Mushroom
from balazar.photo_evaluator import evaluate

import balazar.base as base

KESAKO_OBJ = _("__kesako_obj__")
DROP_OBJ   = _("__drop_obj__")


class Activity(object):
  priority = 0
  next_activity = None
  def can_target(self, user, target): return 0
  def is_default(self, user, target): return 0
  def do(self, user, target): pass
  
  def __cmp__(self, other): return cmp(self.priority, other.priority)
  
class Spell(Activity):
  priority = -10
  effect = 1
  def __init__(self, mana_cost = 0.1):
    self.mana_cost = mana_cost
    
  def do(self, user, target):
    if user.mana < self.mana_cost:
      if user.player and user.queue.local:
        user.queue.controler.show_panel(_("__not_enough_mana__"))
    else:
      casting = self.cast(user, target)
      user.queue.controler.animate_and_cancel(casting)
      return casting
    
  def cast(self, user, target):
    yield State(user, animation = "sort%s" % self.effect)
    for i in range(user.sort_duration1[self.effect]): yield State(user, animation = "")
    yield ActivityState(user, activity = self, target = target)
    for i in range(user.sort_duration2[self.effect]): yield State(user, animation = "")
    
  def casted(self, user, target):
    user.mana -= self.mana_cost

class Kesako(Activity):
  priority = 10
  def can_target(self, user, target): return target
    
  def do(self, user, target):
    if callable(target.kesako): user.queue.controler.show_panel(target.kesako())
    else:                       user.queue.controler.show_panel(target.kesako  )
    
    
class Goto(Activity):
  priority = 10
  def can_target(self, user, target): return target and (user.distance_to(target) > 2.0)
  def do(self, user, target):
    from balazar.controler import Goto
    going = Goto(user, user.queue.controler.location)
    user.queue.controler.append_and_cancel(going)
    return going
    
class Lightening(Spell):
  priority = -15
  effect = 0
  def can_target(self, user, target): return isinstance(target, Character) and (target.alignment * user.alignment < 4)
  def casted(self, user, target):
    Spell.casted(self, user, target)
    LighteningEffect(user.level, user, target).set_xyz(*user.level.transform_point(0.25, 2.0, -1.5, user))
    
class Heal(Spell):
  priority = -15
  def can_target(self, user, target): return isinstance(target, Character)
  def casted(self, user, target):
    Spell.casted(self, user, target)
    target.life = 1.0
    HealEffect(target).y = 5.0
    
class Levitate(Spell):
  def can_target(self, user, target): return target is user
  def cast(self, user, target):
    yield State(user, animation = "sort%s" % self.effect)
    for i in range(user.sort_duration1[self.effect]): yield State(user, animation = "")
    yield ActivityState(user, activity = self, target = target)
    for i in range(user.sort_duration2[self.effect]):
      a = State(user, y = 0.5, animation = "")
      a.unfinished = 1
      user.future_speed_y = 0.5
      yield a
    for i in range(30):
      a = State(user, y = 0.5, animation = "chute")
      a.unfinished = 1
      user.future_speed_y = 0.5
      yield a
    
class Invisibility(Spell):
  def can_target(self, user, target): return isinstance(target, Character)
  def casted(self, user, target):
    user.queue.controler.show_panel("TODO")
    
class ResurrectShield(Spell):
  def can_target(self, user, target): return 0 # Uncastable spell, not treated internally as a spell. This activity exists only for Kesako.
  def cast(self, user, target):
    yield State(user, animation = "mort")
    for i in range(user.die_duration1): yield State(user, z = 0.04, animation = "")
    for i in range(user.die_duration2): yield State(user, animation = "")
    teleportation = user >> Point(user.level, *user.initial_pos)
    teleportation.convert_to(user)
    yield State(user, teleportation.x, teleportation.y, teleportation.z) # => teleport to initial position
    yield ActivityState(user, activity = self, target = target)
    
  def casted(self, user, target):
    Spell.casted(self, user, target)
    target.life = 1.0
    
    
class LightCitadel(Spell):
  priority = -20
  def can_target(self, user, target):
    import balazar.vegetation as vegetation
    return isinstance(target, vegetation.GiantMushroom) and target.state >= 3.0
  def casted(self, user, target):
    user.queue.controler.show_panel("TODO")
    
class Sacrifice(Spell):
  def can_target(self, user, target): return isinstance(target, Item) and ((not target.owner) or (target.owner is user)) and (target.blessed >= 0)
  def casted(self, user, target):
    user.mana += 0.05 + 0.1 * float(target.blessed)
    
    SacrificeEffect(user.level).move(target)
    if target.owner: target.owner.drop_item(target)
    target.parent.remove(target)
    
class Curse(Spell):
  def can_target(self, user, target): return isinstance(target, Character) and (not target is user)
  def casted(self, user, target):
    Spell.casted(self, user, target)
    
    from balazar.item import CursedSword
    cursed_sword = CursedSword(user.level)
    cursed_sword.move(user)
    cursed_sword.owner_name = target.name
    
class Uncurse(Spell):
  priority = -20
  def can_target(self, user, target):
    if isinstance(target, Character):
      for item in target.items:
        if item.blessed < 0: return 1
        
  def casted(self, user, target):
    Spell.casted(self, user, target)
    
    for item in target.items:
      if item.blessed < 0:
        target.drop_item(item)
        item.reset_curse()
        
class ListSpells(Activity):
  priority = -1
  def can_target(self, user, target): return target is user
  def do(self, user, target):
    spells =      map(ListedItem, filter(lambda activity: isinstance(activity, Spell), target.extra_activities))
    spells.extend(map(ListedItem, filter(lambda activity: isinstance(activity, Spell), target.activities)))
    if spells: user.queue.controler.show_menu(spells, user)
    
class ListedSpell:
  def __init__(self, spell):
    self.spell = spell
    self.display_name = KESAKO_OBJ % spell.display_name
  def do(self, user):
    kesako = _("__kesako__%s" % self.spell.__class__.__name__)
    if kesako[0] == "_": kesako = "???"
    user.queue.controler.show_panel(kesako)
    
    
# class ListItems(Activity):
#   priority = 0
#   def can_target(self, user, target): return (target is user) and target.items
#   def do(self, user, target):
#     if target.items: user.queue.controler.show_menu(map(ListedItem, target.items), user)
    
# class ListedItem:
#   def __init__(self, item):
#     self.item = item
#     self.display_name = KESAKO_OBJ % item.display_name
#   def do(self, user):
#     if hasattr(self.item, "kesako"):
#       kesako = self.item.kesako()
#     else:
#       kesako = _("__kesako__%s" % self.item.__class__.__name__)
#       if kesako[0] == "_": kesako = "???"
#     user.queue.controler.show_panel(kesako)
    
    
class DropItem(Activity):
  priority = 5
  def can_target(self, user, target): return isinstance(target, Item) and (user is target.owner)
  def do(self, user, target):
    from balazar.controler import DropItem
    doing = DropItem(user, target)
    user.queue.controler.append_and_cancel(doing)
    return doing
  
# class DropItem(Activity):
#   priority = 5
#   def can_target(self, user, target): return ((target is user) and target.items) or (isinstance(target, Item) and (user is target.owner))
#   def do(self, user, target):
#     if isinstance(target, Item):
#       from balazar.controler import DropItem
#       user.queue.controler.append_and_cancel(DropItem(user, target))
#     else:
#       if target.items: user.queue.controler.show_menu(map(ListedDropItem, target.items), target)
      
# class ListedDropItem:
#   def __init__(self, item):
#     self.item = item
#     self.display_name = DROP_OBJ % item.display_name
#   def do(self, user):
#     from balazar.controler import DropItem
#     user.queue.controler.append_and_cancel(DropItem(user, self.item))
    
class GrabItem(Activity):
  priority = -15
  def can_target(self, user, target): return isinstance(target, Item) and (not target.owner)
  def do(self, user, target):
    from balazar.controler import GrabItem
    doing = GrabItem(user, target)
    user.queue.controler.append_and_cancel(doing)
    return doing
    
class EquipItem(Activity):
  priority = -20
  def can_target(self, user, target): return isinstance(target, EquipableItem) and (not target.equiped) and (target.owner is user)
  def do(self, user, target):
    target.set_equiped(1)
    
class UnequipItem(Activity):
  priority = -20
  def can_target(self, user, target): return isinstance(target, EquipableItem) and target.equiped and (target.owner is user)
  def do(self, user, target):
    target.set_equiped(0)
    
class Telekinesy(Spell):
  effect = 0
  def can_target(self, user, target): return isinstance(target, Item) and not target.owner is user
  def casted(self, user, target):
    Spell.casted(self, user, target)
    TelekinesyEffect(user.level, user, target).move(target)
    if target.owner: target.owner.drop_item(target)
    
    # TODO : agir sur l'IA du perso  qui on a voler l'objet (si il y a vol et si
    # la victime est jou par une IA).
    
class Boreal(Spell):
  def can_target(self, user, target): return target is user
  def casted(self, user, target):
    Spell.casted(self, user, target)
    BorealEffect(user.level)
    
class HandleCamera(Activity):
  priority = -30
  next_activity = 1
  def can_target(self, user, target): return isinstance(target, Camera) and ((target.owner is user)) and target.equiped
  def do(self, user, target):
    doing = self.cast(user, target)
    #user.queue.controler.animate_and_cancel(self.end_casting(user, target))
    user.queue.controler.append_and_cancel(self.end_casting(user, target))
    user.queue.controler.append(doing)
    return doing
  
  def cast(self, user, target):
    if user.queue.local and user.player:
      traveling = CameraTraveling(target)
      soya.IDLER.camera.add_traveling(traveling)
    else:
      traveling = None
      
    for i in range(20):
      s = State(user, animation = "photographie")
      s.unfinished = 1
      yield s
      
    if traveling:
      traveling.smooth_rotation = traveling.smooth_move = 0
      
    while 1:
      soya.IDLER.camera.zap()
      s = State(user, animation = "photographie")
      s.unfinished = 1
      yield s
      
  def end_casting(self, user, target):
    if user.queue.local and user.player:
      for traveling in soya.IDLER.camera.travelings[:]:
        if isinstance(traveling, CameraTraveling): soya.IDLER.camera.remove_traveling(traveling)
        
      soya.IDLER.camera.traveling.direction.move(soya.Vector(user, 0.0, 0.0, 2.0))
      
    s = State(user, animation = "photographie")
    s.unfinished = 1
    yield s
    
class TakePhoto(Activity):
  priority = -20
  def can_target(self, user, target):
    print isinstance(target, Camera), ((target.owner is user)), target.equiped
    return isinstance(target, Camera) and ((target.owner is user)) and target.equiped
  def do(self, user, target):
    doing =  self.cast(user, target)
    user.queue.controler.animate(doing)
    return doing
  
  def cast(self, user, target):
    import balazar.discussion as discussion
    if discussion.CURRENT_DISCUSSION: discussion.CURRENT_DISCUSSION.visible = 0
    user.visible = 0
    yield State(user, animation = "photographie")
    
    s = soya.screenshot(y = soya.root_widget.status_bar.height, height = soya.get_screen_height() - soya.root_widget.status_bar.height)
    photo = Photo(None, s)
    evaluate(user, photo, soya.IDLER.camera)
    user.grab_item(photo)
    
    user.visible = 1
    if discussion.CURRENT_DISCUSSION: discussion.CURRENT_DISCUSSION.visible = 1
    
    effect = PhotoVortex(user.level)
    effect.move(target)
    effect.look_at(soya.Vector(user, 0.0, 0.0, -1.0))
    
    for i in range(20):
      yield State(user, animation = "photographie")
      
      
class PhotoVortex(soya.PythonCoordSyst, base.PhotoScoreable):
  no_close_up = 0
  def get_photo_score(self, photo, camera):
    if self.state > 75.0: return 6.0, 3.0, 5.0
    else:                 return 3.0, 2.0, 5.0

  def display_name(self):
    return _("photo_vortex")
  
  def __init__(self, parent):
    soya.PythonCoordSyst.__init__(self, parent)
    self.state    = 0.0
    self.material = soya.Material.get("fx_photo")
    #self.material.additive_blending = 1
    #self.material.save()
    
  def destroy(self):
    if self.parent: self.parent.remove(self)
    
  def advance_time(self, proportion):
    if self.state > 75.0:
      soya.IDLER.next_round_tasks.append(self.destroy)
    else:
      self.state += proportion
      
  def batch(self):
    return 2, self, self.material
  
  def render(self):
    import soya.opengl as opengl
    
    self.material.activate()
    
    opengl.glDisable(opengl.GL_CULL_FACE)
    opengl.glDisable(opengl.GL_LIGHTING)
    opengl.glDisable(opengl.GL_FOG)
    opengl.glBegin(opengl.GL_TRIANGLES)
    
    angle = (-10.0) * 0.017453292519943295
    ac    = math.cos(angle)
    as    = math.sin(angle)
    h     = math.sin(self.state * 0.062831853071795868) * 10.0
    
    color = 1
    for i in range(1):
      for angle in range(0, 360, 10):
        angle = (angle) * 0.017453292519943295
        c = math.cos(angle)
        s = math.sin(angle)
        z1 = - i      * 15.0 - min(0, h)
        z2 = -(i + 1) * 15.0 - h
        
        opengl.glColor4f(0.3, 0.7, 1.0, 1.0 - self.state / 75.0)
        opengl.glTexCoord2f(0.5, 1.0); opengl.glVertex3f(c  * z1, s  * z1, z1)
        
        opengl.glColor4f(1.0, 1.0, 1.0, 0.0)
        opengl.glTexCoord2f(1.0, 0.04); opengl.glVertex3f(ac * z2, as * z2, z2)
        opengl.glTexCoord2f(0.0, 0.04); opengl.glVertex3f(c  * z2, s  * z2, z2)
        
        color = not color
        ac = c
        as = s
        
    opengl.glEnd()
    opengl.glColor4f(1.0, 1.0, 1.0, 1.0)
    opengl.glEnable(opengl.GL_CULL_FACE)
    opengl.glEnable(opengl.GL_LIGHTING)
    opengl.glEnable(opengl.GL_FOG)
    
class CameraTraveling(soya.FixTraveling):
  def __init__(self, camera):
    self.camera = camera
    soya.FixTraveling.__init__(self, soya.Point(camera.owner, 0.3, 2.5, 0.1), soya.Point(camera.owner, 0.3, 2.5, -1.0), 1, 1)
    

class UseItem(Activity):
  priority = -20
  def can_target(self, user, target): return isinstance(target, UseableItem) and ((target.owner is user))
  def do(self, user, target):
    doing = self.cast(user, target)
    user.queue.controler.animate_and_cancel(doing)
    return doing
  
  def cast(self, user, target):
    yield UseItemState(user, item = target, activity = self)
  def casted(self, user, item):
    item.used_by(user)
    
class ViewItem(UseItem):
  pass

class EatItem(UseItem):
  priority = -20
  def can_target(self, user, target): return isinstance(target, Food) and ((target.owner is user))
  def casted(self, user, item):
    item.eaten_by(user)
    
class Attack(Activity):
  priority = 0
  def can_target(self, user, target): return isinstance(target, Character) and (not target is user)
  def do(self, user, target):
    if not user.queue.controler.is_enemy(target):
      enemies = user.queue.controler.enemies
      enemies.append(target.name)
      if len(enemies) > 8: del enemies[0]
    doing = self.cast(user, target)
    user.queue.controler.append_and_cancel(doing)
    return doing
  
  def cast(self, user, target):
    from balazar.controler import Goto, LookAt
    while target.life > 0.0:
      yield Goto  (user, target, 1.7)
      yield LookAt(user, target)
      for i in range(20): yield State(user)
      
class DontAttack(Activity):
  priority = 0
  def can_target(self, user, target): return isinstance(target, Character) and (not target is user) and target.name in user.queue.controler.enemies
  def do(self, user, target):
    enemies = user.queue.controler.enemies
    try: enemies.remove(target.name)
    except ValueError: pass
    

    
    
def _compute_names(clazz):
  for subclass in clazz.__subclasses__(): _compute_names(subclass)
  if not hasattr(clazz, "display_name"):
    clazz.display_name = _("__activity__%s" % clazz.__name__)
_compute_names(Activity)


# Defines objects's activities
Item.target_activities = [
  GrabItem(),
  DropItem(),
  ]
EquipableItem.target_activities = [
  EquipItem(),
  UnequipItem(),
  ] + Item.target_activities
UseableItem.target_activities = [
  UseItem(),
  ] + Item.target_activities
Photo.target_activities = [
  ViewItem(),
  ] + Item.target_activities
Food.target_activities = [
  EatItem(),
  ] + Item.target_activities
Camera.target_activities = [
  HandleCamera(),
  TakePhoto(),
  ] + EquipableItem.target_activities




import random

class SpellEffect: pass

class HealEffect(soya.Particles, SpellEffect):
  def __init__(self, parent = None, nb_particles = 22):
    soya.Particles.__init__(self, parent, soya.PARTICLE_DEFAULT_MATERIAL, nb_particles)
    self.auto_generate_particle = 1
    self.set_colors((1.0, 0.2, 0.8, 0.4), (1.0, 0.2, 0.8, 1.0), (1.0, 0.2, 0.2, 1.0), (0.8, 0.2, 0.1, 1.0), (0.3, 0.1, 0.0, 1.0))
    self.set_sizes((1.0, 1.0))
    self.life = 40
    self.max_particles_per_round = 4
    
  def generate(self, index):
    angle = random.uniform(0.0, 6.2834)
    sx = math.cos(angle)
    sy = -7.0
    sz = math.sin(angle)
    l = 0.15 / math.sqrt(sx * sx + sy * sy + sz * sz)
    self.set_particle(index, random.uniform(0.6, 1.5), sx * l, sy * l, sz * l, 0.0, 0.0, 0.0)
    
  def begin_round(self):
    soya.Particles.begin_round(self)
    if   self.life >  0: self.life -= 1
    elif self.life == 0:
      self.auto_generate_particle = 0
      self.removable = 1
      self.life = -1
      
class LighteningEffect(soya.Particles, SpellEffect):
  def __init__(self, parent = None, user = None, target = None, nb_particles = 60):
    soya.Particles.__init__(self, parent, soya.Material.get("x_cristal_1"), nb_particles)
    self.auto_generate_particle = 1
    self.set_colors((0.1, 0.3, 0.6, 1.0), (0.2, 0.6, 1.0, 1.0), (0.1, 0.6, 1.0, 1.0), (0.0, 0.1, 0.3, 1.0))
    self.set_sizes((0.8, 0.8))
    self.life = 1000
    self.max_particles_per_round = 4
    self.user   = user
    self.target = target
    self.target_p = Point(target, 0.0, 1.2, 0.0)
    self.speed = Vector(user, 0.0, 0.0, -0.3)
    self.speed.convert_to(parent)
    self.v = Vector()
    
  def generate(self, index):
    sx = random.uniform(-1.0, 1.0)
    sy = random.uniform(-1.0, 1.0)
    sz = random.uniform(-1.0, 1.0)
    if self.life > 100: l = 0.06 / math.sqrt(sx * sx + sy * sy + sz * sz)
    else:               l = 0.2 / math.sqrt(sx * sx + sy * sy + sz * sz)
    self.set_particle(index, random.uniform(0.5, 1.5), sx * l, sy * l, sz * l, 0.0, 0.0, 0.0)
    
  def begin_round(self):
    soya.Particles.begin_round(self)
    
    if self.life > 100:
      self.v.set_start_end(self, self.target_p)
      self.v.convert_to(self.parent)
      if self.v.length() < 1.0:
        self.life = 20
        self.target.hurt(1.0, self.user)
        
      else:
        self.v.set_length(0.02)
        self.speed += self.v
        self.speed *= (0.3 / 0.32)
        
    elif self.life == 0:
      self.auto_generate_particle = 0
      self.removable = 1
      self.life = -1
      
    else: self.life -= 1
      
  def advance_time(self, proportion):
    soya.Particles.advance_time(self, proportion)
    
    if self.life > 100: self.add_mul_vector(proportion, self.speed)
    
class TelekinesyEffect(soya.Particles, SpellEffect):
  def __init__(self, parent = None, user = None, target = None, nb_particles = 25):
    soya.Particles.__init__(self, parent, soya.Material.get("x_solar_2"), nb_particles)
    self.auto_generate_particle = 1
    self.set_colors((0.1, 0.1, 0.6, 1.0), (0.2, 0.2, 1.0, 1.0), (0.1, 0.1, 1.0, 1.0), (0.0, 0.0, 0.3, 1.0))
    self.set_sizes((0.4, 0.4))
    self.max_particles_per_round = 1
    self.user   = user
    self.target = target
    self.target_p = Point(user, 0.6, 1.3, -0.5)
    self.speed = Vector(parent)
    self.v = Vector()
    
    #target.solid = 0
    
  def generate(self, index):
    sx = random.uniform(-1.0, 1.0)
    sy = random.uniform(-1.0, 1.0)
    sz = random.uniform(-1.0, 1.0)
    l = 0.03 / math.sqrt(sx * sx + sy * sy + sz * sz)
    self.set_particle(index, random.uniform(0.5, 1.5), sx * l, sy * l, sz * l, 0.0, 0.0, 0.0)
    
  def begin_round(self):
    soya.Particles.begin_round(self)
    
    self.v.set_start_end(self, self.target_p)
    self.v.convert_to(self.parent)
    if (self.v.length() < 0.03) and (self.user):
      self.auto_generate_particle = 0
      self.removable = 1
      self.target.solid = 1
      self.user.grab_item(self.target)
      self.user = None
    else:
      self.v.set_length(0.02)
      self.speed += self.v
      self.speed *= (0.2 / 0.22)
      
  def advance_time(self, proportion):
    soya.Particles.advance_time(self, proportion)
    
    self.add_mul_vector(proportion, self.speed)
    if not self.target.owner: self.target.move(self)
    
class MentalDeathEffect(soya.Particles, SpellEffect):
  def __init__(self, parent = None, nb_particles = 15):
    soya.Particles.__init__(self, parent, soya.PARTICLE_DEFAULT_MATERIAL, nb_particles)
    self.auto_generate_particle = 1
    self.set_colors((0.0, 0.0, 0.5, 1.0), (0.1, 0.1, 1.0, 1.0), (0.3, 0.2, 1.0, 1.0), (0.1, 0.0, 0.3, 1.0))
    self.set_sizes((1.5, 1.5))
    self.life = 15
    self.max_particles_per_round = 8
    
  def generate(self, index):
    angle = random.uniform(0.0, 6.2834)
    sx = math.cos(angle)
    sz = math.sin(angle)
    l = 0.02 / math.sqrt(sx * sx + 1.0 + sz * sz)
    self.set_particle(index, random.uniform(0.5, 1.5), sx * l, l, sz * l, 0.0, 0.0, 0.0)
    
  def begin_round(self):
    self.set_sizes((self.life / 10.0, self.life / 10.0))
    soya.Particles.begin_round(self)
    if   self.life >  1: self.life -= 1
    elif self.life == 1:
      self.auto_generate_particle = 0
      self.removable = 1
      self.life = -1
    
class SacrificeEffect(soya.Particles, SpellEffect):
  def __init__(self, parent = None, nb_particles = 25):
    soya.Particles.__init__(self, parent, soya.PARTICLE_DEFAULT_MATERIAL, nb_particles)
    self.auto_generate_particle = 1
    self.set_colors((1.0, 0.0, 0.0, 1.0), (1.0, 0.0, 0.0, 1.0), (0.3, 0.0, 0.0, 1.0))
    self.set_sizes((1.0, 1.0))
    self.life = 15
    self.max_particles_per_round = 8
    
  def generate(self, index):
    sx = random.uniform(-1.0, 1.0)
    sy = random.uniform(-1.0, 1.0)
    sz = random.uniform(-1.0, 1.0)
    l = 0.06 / math.sqrt(sx * sx + sy * sy + sz * sz)
    self.set_particle(index, random.uniform(0.5, 1.5), sx * l, sy * l, sz * l, 0.0, 0.0, 0.0)
    
  def begin_round(self):
    soya.Particles.begin_round(self)
    if   self.life >  1: self.life -= 1
    elif self.life == 1:
      self.auto_generate_particle = 0
      self.removable = 1
      self.life = -1
      
      
class BorealEffect(soya.Volume):
  def __init__(self, parent = None):
    soya.Volume.__init__(self, parent)
    
    self.nb = 0
    
  def begin_round(self):
    if self.parent:
      if self.nb > 120:
        self.parent.remove(self)
      else:
        self.nb += 1
        size = 1.0 + float(self.nb) / 5.0
        day_night = self.parent.atmosphere
        
        for star in day_night.astral_world.children[1:]:
          if isinstance(star, soya.Sprite) and hasattr(star, "size"):
            star.width = star.height = star.size * size
            
