# Balazar
# Copyright (C) 2003 Jean-Baptiste LAMY
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import math, random
import soya
from soya import Point, Vector

from balazar.controler  import StackControler, Goto, Gotoward, LookAt, GrabItem, Fly
from balazar.character  import State, Character
from balazar.hero       import Balazar, Beuglaz
from balazar.item       import Item, Food, Key
from balazar.vegetation import Mushroom
from balazar.activity   import SpellEffect


class DretzelCandidateAI(StackControler):
  vision       = 50.0
  check_period = 80
  aggressivity = 0.0
  
  def __init__(self, character = None, residence = None):
    StackControler.__init__(self, character)
    
    self.residence  = residence
    self.relations  = {}
    self.check_time = 80
    
    self.append(self.take_decision())

  def __setstate__(self, state):
    self.__dict__ = state
    self.append(self.take_decision())
    
  def item_given(self, item, by):
    self.relations[by.name] = self.get_relation(by) + self.interest(item, 1.0) / 2.0
    
  def item_gained(self, item):
    if isinstance(item, Food):
      if self.character.hunger > 0.2: # Eat
        item.eaten_by(self.character)
        
  def is_enemy(self, character):
    #(self.character.alignment * character.alignment < 0)
    return self.get_relation(character) < -0.1
  
  def next(self):
    if not self.uncancelable_controler:
      if self.check_time == 0:
        self.check_time = self.check_period
        self.check()
      else: self.check_time -= 1
      
    state = StackControler.next(self)
    
    #state = State(self.character)
    #state.x = state.y = state.z = state.rotation_speed = 0.0
    #state.animation = "attente"
    if state.x > -1000.0: state = self.character.collision_detection(state)
    return state
  
  def check(self):
    interests = map(lambda o: (self.interest(o), o), self.character.level)
    a = max(interests)
    b = min(interests)
    
    if a[0] > -b[0]: o = a
    else:            o = b
    if abs(o[0]) > 0.2: self.append(self.deal_with(*o))

    #interests.sort()
    #print "CHECK", interests
    
    
  def interest(self, o, d = None):
    """Returns a value in the range [-1.0, 1.0] indicating if the AI is interested
by O (where 1.0 means very interested, 0.0 means ignore it, and -1.0 means go away
from it)."""
    if o is self.character: return 0.0
    if d is None: d = 1.0 + self.character.distance_to(o) / self.vision
    
    if isinstance(o, Item):
      incomp = 0.0
      if not o.owner is self.character:
        for item in self.character.items:
          if item.incompatible_with(o): incomp += self.interest(item)
      if isinstance(o, Food):
        if self.character.qualities.has_key("ascetic"): return 0.0
        if self.character.qualities.has_key("greedy"):  return (1.0 - incomp) / d
        return self.character.hunger
      return (0.1 + o.blessed - incomp) / 3.0
    if isinstance(o, Character): return min(0.0, self.get_relation(o) / d)
    if isinstance(o, SpellEffect):
      if self.character.qualities.has_key("hatesorcerer"): return -0.8 / d
      if self.character.qualities.has_key("lovemagic"):    return  0.3 / d
      
    return 0.0 # Unidentifiable or useless thing
  
  def get_relation(self, character):
    relation = self.relations.get(character.name, None)
    if relation is None:
      if isinstance(character, Balazar):
        if self.character.qualities.has_key("hatesorcerer"): return -0.5
        if self.character.qualities.has_key("lovemagic"):    return  0.3
      if character.alignment * self.character.alignment < 0: return -0.3
      return 0.0
    return relation
  
  def change_relation(self, character, offset):
    self.relations[character.name] = min(1.0, max(-1.0, self.get_relation(character) + offset))
    
  def deal_with(self, interest, o):
    if interest < 0.0:
      if isinstance(o, Character): yield self.attack(o)
      else:                        yield Fly     (self.character, o)
    elif isinstance(o, Item):      yield GrabItem(self.character, o)
    else:                          yield Goto    (self.character, o)
    
  def attack(self, ennemy):
    while ennemy.life > 0.0:
      a = self.aggressivity + self.character.life - ennemy.life
      
      if   self.character.qualities.has_key("brave" ): a += 0.5
      elif self.character.qualities.has_key("coward"): a -= 2.0
      
      if a > 0.0:
        yield Goto(self.character, ennemy, 1.5)
        yield LookAt(self.character, ennemy)
        for i in range(50): yield State(self.character)
      else:
        yield Fly(self.character, ennemy)
        break
      
  def go_for_food(self):
    for item in self.character.items:
      if isinstance(item, Food):
        yield(item.target_activities[2].cast(self.character, item))
    else:
      pass
      # XXX Aller au marche, prendre  bouffer, bouffer,...
      
  def go_to_bed(self):
    self.append(self.residence.Goto(self))
    # XXX S'allonger, dormir.
  
  
class FarmerAI(DretzelCandidateAI):
  aggressivity  = 0.2
  
  def __init__(self, character):
    DretzelCandidateAI.__init__(self, character)
    self.task = ""
    
  def interest(self, o, d = None):
    if d is None: d = 1.0 + self.character.distance_to(o) / self.vision
    if d > self.vision: return 0.0
    
    if isinstance(o, Item):
      incomp = 0.0
      if not o.owner is self.character:
        for item in self.character.items:
          if item.incompatible_with(o): incomp += self.interest(item)
      if isinstance(o, Food): return (1.0 - incomp) / d
      if isinstance(o, Key ): return (0.3 - incomp) / d
      
    if isinstance(o, Mushroom):
      if self.task == "work":
        for item in self.character.items:
          if isinstance(item, Food): return 0.0
        if o.state >= 2.0: return 0.5 / d # Old mushroom
        else:              return 0.8 / d
      return 0.0
    
    return DretzelCandidateAI.interest(self, o, d)
    
  def deal_with(self, interest, o):
    if interest > 0.0:
      if isinstance(o, Mushroom):
        yield Goto(self.character, o)
        yield LookAt(self.character, o)
        self.uncancelable_controler = self.character.activities[0].cast(self.character, o)
        yield self.uncancelable_controler
        
        
        #if self.distance_to(o) < 1.0:
        #  for i in range(330):
        #    yield State(self.character, animation = "travail")
        
        return
    yield DretzelCandidateAI.deal_with(self, interest, o)
    
  def take_decision(self):
    while 1:
      #if self.character.hunger > 0.4: self.go_for_food()
      #else:
      
      self.task = "work"
      for i in range(10):
        for i in range(20):
          p = Point(self.character, random.uniform(-15.0, 15.0), 1000.0, random.uniform(-20.0, 5.0))
          p.convert_to(self.character.level)
          r = self.character.level.raypick(p, self.character.down, -1.0, 1, 1)
          if r:
            p.y = r[0].y
            yield Gotoward(self.character, p, 50)
            break
        for i in range(50): yield State(self.character)



