# Balazar
# Copyright (C) 2003 Jean-Baptiste LAMY
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import sys, os, os.path, math, cPickle as pickle
import soya
from soya import Point, Vector

import balazar.base as base, balazar.sound as sound
from balazar.base import Strikeable
from balazar.item import KillRay, Item, Weapon

import struct
from socket import MSG_WAITALL, error as SocketError


class State(object):
#   def __new__(clazz, character = None, x = 0.0, y = 0.0, z = 0.0, rotation_speed = 0, animation = "attente", auto_jump = 0):
#     if (not isinstance(character, int)) and character and character.reuseable_state:
#       s = character.reuseable_state
#       return s
#     else:
#       return object.__new__(clazz, character, x, y, z, rotation_speed, animation, auto_jump)
    
  def __init__(self, character = None, x = 0.0, y = 0.0, z = 0.0, rotation_speed = 0, animation = "attente", auto_jump = 0):
    self.character      = character
    self.x              = x
    self.y              = y
    self.z              = z
    self.rotation_speed = rotation_speed
    self.animation      = animation
    self.auto_jump      = auto_jump
    self.unfinished     = 0
    
  def __repr__(self):
    return "<%s %s (%s %s %s, %s)>" % (self.__class__.__name__, self.animation, self.x, self.y, self.z, self.rotation_speed)
  
  def apply(self): pass
  
class DieState(State):
  def apply(self): self.character.die()
  
class ItemState(State):
  def __init__(self, character = None, x = 0.0, y = 0.0, z = 0.0, rotation_speed = 0, animation = "attente", auto_jump = 0, item = None):
    self.item = item
    State.__init__(self, character, x, y, z, rotation_speed, animation, auto_jump)
    
class DropItemState(ItemState):
  def apply(self):
    self.character.drop_item(self.item)

class GrabItemState(ItemState):
  def apply(self):
    self.character.grab_item(self.item)

class UseItemState(ItemState):
  def __init__(self, character = None, x = 0.0, y = 0.0, z = 0.0, rotation_speed = 0, animation = "attente", auto_jump = 0, item = None, activity = None):
    self.activity = activity
    ItemState.__init__(self, character, x, y, z, rotation_speed, animation, auto_jump, item)
  def apply(self):
    self.activity.casted(self.character, self.item)
    
class ActivityState(State):
  def __init__(self, character = None, x = 0.0, y = 0.0, z = 0.0, rotation_speed = 0, animation = "attente", auto_jump = 0, activity = None, target = None):
    self.activity = activity
    self.target   = target
    State.__init__(self, character, x, y, z, rotation_speed, animation, auto_jump)
  def apply(self):
    self.activity.casted(self.character, self.target)
    
class FightState(State):
  def __init__(self, character = None, x = 0.0, y = 0.0, z = 0.0, rotation_speed = 0, animation = "attente", auto_jump = 0, kill_rays = None):
    self.kill_rays = kill_rays or (character.weapon and character.weapon.kill_rays) or character.hand_kill_rays
    State.__init__(self, character, x, y, z, rotation_speed, animation, auto_jump)
    
  def apply(self):
    self.character.kill_rays = self.kill_rays
    
class FightHitState(FightState):
  def __init__(self, character = None, x = 0.0, y = 0.0, z = 0.0, rotation_speed = 0, animation = "attente", auto_jump = 0, kill_rays = None, target = None, damage = 1):
    self.target = target
    self.damage = damage
    FightState.__init__(self, character, x, y, z, rotation_speed, animation, auto_jump, kill_rays)
    self.unfinished = 1
    
  def apply(self):
    FightState.apply(self)
    if not self.target.hurt(self.character, self.damage):
      self.character.queue.controler.animate_and_cancel(WeaponHurted(self.character))
    
    

PRIORITARY_ANIMATIONS = { # Defaults to 1
  "attente" : 0,
  "marche"  : 0,
  "recule"  : 0,
  "tourneG" : 0,
  "tourneD" : 0,
  "chute"   : 0,
  }

_P  = Point()
_V  = Vector()
_P2 = Point()
_V2 = Vector()

_TEMP = [
  (_P , _V ),
  (_P2, _V2),
  ]

_CONTEXT = None

class LocalQueue(list):
  local = 1
  def __init__(self, character):
    self.character = character
    self.controler = None

  def pop(self):
    self.append(self.controler.next())
    return list.pop(self, 0)
  
  
class Character(soya.World, base.Entity):
  mana        = 0.0
  alignment   = 0.0
  tolerance   = 0.5
  
  move_speed  = 1.0
  radius      = 0.5
  radius_y    = 1.2
  range       = 0.8
  
  def __init__(self, player):
    soya.World.__init__(self)
    
    self.player = player
    self.id     = 0
    self.level  = None
    self.queue  = LocalQueue(self)
    
    self.perso = soya.Cal3dVolume(self)
    
    self.current_animation = ""
    self.speed             = Vector(None)
    self.rotation_speed    = 0.0
    self.center            = Point(self, 0.0, self.radius_y, 0.0)
    self.jumping           = 0
    self.lod               = 0
    self.ground            = Point()
    self.ground_normal     = Vector()
    self.last_ground       = Point()
    self.on_ground         = 0
    self.life              = 1.0
    self.future_speed_y    = 0.0
    self.fight_round       = 0
    self.invincibility     = 0
    
    self.fighting   = 2.0
    self.resistance = 2.0
    
    self.left   = Vector(self, -1.0,  0.0,  0.0)
    self.right  = Vector(self,  1.0,  0.0,  0.0)
    self.down   = Vector(self,  0.0, -1.0,  0.0)
    self.up     = Vector(self,  0.0,  1.0,  0.0)
    self.front  = Vector(self,  0.0,  0.0, -1.0)
    self.back   = Vector(self,  0.0,  0.0,  1.0)
    
    self.items           = []
    self.weapon          = None
    self.kill_rays       = None
    self.left_hand_item  = None
    self.reuseable_state = None
    self.name            = base.unique()
    
    self.relations  = {}
    
    self.solid      = 0

  def get_delta_alignment(self, character):
    r = self.alignment - character.alignment
    if   r >  0.3: return 0
    elif r < -0.3: return 2
    return 1
  
  def is_enemy(self, character):
    #(self.character.alignment * character.alignment < 0)
    return self.get_relation(character) < 0.0
  
  def get_relation(self, character):
    relation = self.relations.get(character.name, None)
    if relation is None:
      relation = character.relations.get(self.name, None)
      if relation is None:
        return (character.tolerance + self.tolerance) - abs(character.alignment - self.alignment)
    return relation
  
  def change_relation(self, character, offset):
    self.relations[character.name] = self.get_relation(character) + offset
    self     .queue.controler.relation_changed(character)
    character.queue.controler.relation_changed(self     )
    
  def init_kill_rays(self):
    self.right_hand = soya.World(self)
    self.perso.attach_to_bone(self.right_hand, "mainD")
    self.hand_kill_rays = [
      KillRay(self.right_hand, 0.0, 0.2, 0.0, 0.0, 0.0, 0.5),
      ]
    
  def get_strike_rot(self, target):
    if target:
      _P.clone(target)
      _P.convert_to(self)
      if   _P.x < -0.2: return  4.0
      elif _P.x >  0.2: return -4.0
    return 0.0
    
  def grab_item(self, item):
    if item.parent: item.parent.remove(item)
    self.items.append(item)
    item.set_owner(self)
    item.zap()
    
  def drop_item(self, item):
    item.set_owner(None)
    if item in self.items:
      self.items.remove(item)
      self.level.add(item)

      self.center.z -= 0.7
      r = self.level.raypick(soya.Point(self, 0.0, 2.0, -1.5), self.down, -1.0, 1, 1)
      self.center.z += 0.7

      item.set_identity()
      if r:
        r[0].convert_to(item.parent)
        item.set_xyz(r[0].x, r[0].y + 0.05, r[0].z)
        item.look_at(self.front)
        item.look_at_y(r[1])
      else:
        item.set_xyz(self.x, self.y + 0.05, self.z)
        item.look_at(self.front)
      if item.can_rotate: item.turn_incline(90.0)
      item.zap()

  def item_of_class(self, itemclazz):
    for item in self.items:
      if isinstance(item, itemclazz): return item
      
  def remove_item_of_class(self, itemclazz):
    for item in self.items:
      if isinstance(item, itemclazz):
        item.set_owner(None)
        self.items.remove(item)
        return item
      
  def zap(self):
    for item in self.items: item.zap()
    
  def set_level(self, level):
    if self.player and self.queue.local:
      if self.level: soya.IDLER.scene.remove(self.level)
      
      
    if self.level:
      self.level.remove_character(self)
      if self.player and self.queue.local: self.level.set_active(0)
    self.level = level
    if level:
      level.add_character(self)
      if self.player and self.queue.local: level.set_active(1)
      
    # ???
    self.new_center = Point(soya.IDLER.scene, self.x, self.y + self.radius_y, self.z)
    self.last_state_x = self.x
    self.last_state_z = self.z
    
    if self.player and self.queue.local:
      if level: soya.IDLER.scene.insert(0, level)
      
      
  def collision_detection(self, state):
    """Checks for collision and modify state.x, state.y and state.z if needed.
Also converts state.x, state.y and state.z from the character coordinate system
to level coordinate system."""
    old_x = self.new_center.x
    old_y = self.new_center.y - self.radius_y
    old_z = self.new_center.z
    
    if state.y > 0.0: # Jump
      if self.on_ground:
        self.on_ground = 0
        self.jumping = 1
        self.future_speed_y = state.y
        if state.y > 0.3: sound.play("jump1.wav", self)
        
      else: state.y = self.future_speed_y
    else:
      if state.y < 0.0:
        if self.future_speed_y > 0.0: self.future_speed_y = 0.0
      state.y += self.future_speed_y
    
    # For mobile platform
    if self.on_ground == 2:
      self.ground.convert_to(self.level)
      self.new_center.add_xyz(self.ground.x - self.last_ground.x, self.ground.y - self.last_ground.y, self.ground.z - self.last_ground.z)
      
    self.new_center.add_xyz(*self.parent.transform_vector(state.x, state.y, state.z, self))
    
    #self.solid = 0
    scene = self.get_root()
    
    # Character-character collision
#     for character in self.level.characters.itervalues():
#       if (not character is self) and character.visible:
#         d = character.distance_to(self) - self.radius - character.radius
#         if d < 0.0:
#           self.new_center.add_xyz(d * (character.x - self.x), d * (character.y - self.y), d * (character.z - self.z))
#         if (d <= 2.0) and self.fighting and (self.alignment * character.alignment < 0) and (not ONE_SHOT_ANIMATIONS.get(state.animation, 1)):
#           xyz = self.transform_point(character.x, charcater.y, character.z, character.parent)
#           if self.zone_fight or ((xyz[2] < 0.0) and (-1.0 < xyz[0] / xyz[2] < 1.0)):
#             if self.current_animation == "combat": state.animation = character.name
#             else:                                  state.animation = "combat"

    if self.invincibility: self.invincibility -= 1
    else:
      # XXX raypicking should now do that ???
      
      d = int(self.new_center.x // 4.0) + 48 * int(self.new_center.z // 4.0)
      dx = cmp(self.new_center.x % 4.0, 2.0)
      dz = cmp(self.new_center.z % 4.0, 2.0) * 48
      for d in (d, d + dx, d + dz, d + dx + dz):
        if 0 <= d < 2304:
          characters = self.level.char_pos[d]
          for character in characters:
            if isinstance(character, Character) and (not character is self) and character.visible:
              d = character.distance_to(self) - 1.0 - self.radius - character.radius
              if d < 0.0:
                self.new_center.add_xyz(d * (character.x - self.x), d * (character.y - self.y), d * (character.z - self.z))
#               if character.kill_point:
#                 print character.kill_point.distance_to(self.center), character.kill_point.position()
                
#                 if character.kill_point.distance_to(self.center) < character.kill_point.radius:
#                   fighting = character.fighting
#                   for item in character.items: fighting += item.fighting_bonus(self)
#                   self.hurt(fighting / 10.0, character, character >> self)
#                   self.invincibility = 20
#                   Blood(self.level).move(character.kill_point)
                  
#          if isinstance(state, FightState):
#             for e in characters:
#               if isinstance(e, Strikeable) and state.kill_point.distance_to(e.center) < state.kill_point.radius + e.radius:
#                 fighting = self.fighting - e.resistance
#                 for item in self.items: fighting += item.fighting_bonus(e)
                
#                 self.queue.controler.animate(FightHitState(self, animation = state.animation, target = e, damage = fighting + random.uniform(-10.0, 10.0)))
#                 break
                
      if isinstance(state, FightState):
        characters = self.level.characters.values()
        for character in characters:
          character.solid = 1 # Temporarily enable raypick on characters
          if character.weapon and character.weapon.prot: character.weapon.prot.solid = 1 # Activate raypicking on protection shape
        self.solid = 0 # but not on self !
        
        kill_context = scene.RaypickContext(state.kill_rays[0], self.range, _CONTEXT)
        res = [ kill_context.raypick(kill_ray, kill_ray.dir, kill_ray.length, 1, 0, _TEMP[i][0], _TEMP[i][1])
                for (i, kill_ray) in enumerate(state.kill_rays)
                ]
        for i in range(len(res)):
          if res[i]:
            target = _TEMP[i][0].parent
            while target:
              if   isinstance(target, Weapon):
                if target.owner and not target.owner.invincibility:
                  sound.play("weapon_hurted2.wav", _TEMP[i][0])
                  eclat = Eclat(self.level)
                  eclat.move   (_TEMP[i][0])
                  eclat.look_at(_TEMP[i][1])
                  self.queue.controler.animate_and_cancel(WeaponHurted(self))
                break
              
              target = target.parent
            else: continue
            break
          
        else: # No weapon hurted => check hit
          for i in range(len(res)):
            if res[i]:
              target = _TEMP[i][0].parent
              while 1:
                if   isinstance(target, Character):
                  if (not target.invincibility) and self.is_enemy(target):
                    fighting = self.fighting
                    for item in self.items: fighting += item.fighting_bonus(target)
                    _TEMP[i][1].__imul__(-1)
                    target.hurt(fighting / 10.0, self, *_TEMP[i])
                    target.invincibility = 20
                  break
                
                elif isinstance(target, Strikeable):
                  fighting = self.fighting - target.resistance
                  for item in self.items: fighting += item.fighting_bonus(target)
                  self.queue.controler.animate(FightHitState(self, animation = state.animation, target = target, damage = fighting + random.uniform(-10.0, 10.0)))
                  break
                
                elif isinstance(target, soya.Land): break
                
                elif target is None:
                  if self.weapon and (kill_ray in self.weapon.kill_rays):
                    sound.play("weapon_hurted3.wav", _TEMP[i][0])
                    eclat = Eclat(self.level)
                    eclat.move   (_TEMP[i][0])
                    eclat.look_at(_TEMP[i][1])
                    self.queue.controler.animate_and_cancel(WeaponHurted(self))
                  break
                
                target = target.parent
              else: continue
              break
                
        for character in characters:
          character.solid = 0
          if character.weapon and character.weapon.prot: character.weapon.prot.solid = 0
          
    # Character-level collision
    context = scene.RaypickContext(self.new_center, self.radius_y, _CONTEXT)
    
    if context.raypick(self.new_center, self.left, self.radius, 1, 0, _P, _V):
      _V.convert_to(self.parent)
      _V.normalize()
      self.new_center.add_mul_vector(self.radius - self.new_center.distance_to(_P), _V)

    if context.raypick(self.new_center, self.front, self.radius, 1, 0, _P, _V):
      _V.convert_to(self.parent)
      _V.normalize()
      self.new_center.add_mul_vector(self.radius - self.new_center.distance_to(_P), _V)
      
    if context.raypick(self.new_center, self.up, self.radius_y, 1, 1, _P, _V):
      _V.convert_to(self.parent)
      _V.normalize()
      self.new_center.add_mul_vector(self.radius_y - self.new_center.distance_to(_P), _V)
      if self.future_speed_y > 0.0: self.future_speed_y = 0.0
      
      
    # Fall
    if context.raypick(self.new_center, self.down, self.radius_y, 1, 1, self.ground, self.ground_normal) and not self.jumping:
      if (not self.on_ground) and (self.future_speed_y < -0.15): sound.play("chute.wav", self)
      if self.ground.parent.parent is self.level.children[0]: # Not on a mobile platform (level.children[0] is the static part !!!)
        self.on_ground = 1
        self.new_center.y = self.level.transform_point(self.ground.x, self.ground.y, self.ground.z, self.ground.parent)[1] + self.radius_y
        
      else: # Probably (not sure) on a mobile platform
        self.on_ground = 2
        self.last_ground.__init__(self.level, *self.level.transform_point(self.ground.x, self.ground.y, self.ground.z, self.ground.parent))
        self.new_center.y = self.last_ground.y + self.radius_y
        
    else:
      # Don't interrupt one shot animation such as spellcasting
      if not PRIORITARY_ANIMATIONS.get(state.animation, 1): state.animation = "chute"
      
      if state.auto_jump and self.on_ground:
        self.jumping = 1
        self.future_speed_y = 0.5
      else:
        if self.future_speed_y > -0.25:
          self.future_speed_y -= 0.02
          if self.jumping and (self.future_speed_y < 0.0): self.jumping = 0
        if self.new_center.y < -30.0:
          self.new_center.y = 110.0
          
      self.on_ground = 0
      
    #self.solid = 1
    
    state.x = self.new_center.x
    state.y = self.new_center.y - self.radius_y
    state.z = self.new_center.z
    
    if (state.x == old_x) and (state.y == old_y) and (state.z == old_z):
      if (state.animation == "attente") and (state.__class__ is State):
        self.reuseable_state = state
        return None
      
    return state
    
  def begin_round(self):
    state = self.queue.pop()
    self.set_next_state(state)
    
    soya.World.begin_round(self)
    
  def set_next_state(self, state):
    if not state:
      self.speed.set_xyz(0.0, 0.0, 0.0)
      self.rotation_speed = 0.0
      if self.current_animation != "attente":
        if self.current_animation: self.perso.animate_clear_cycle(self.current_animation)
        self.current_animation = "attente"
        self.perso.animate_blend_cycle("attente")
        
    else:
      state.apply()
      
      self.rotation_speed = float(state.rotation_speed)
      
      ad = int(self.last_state_x // 4.0) + 48 * int(self.last_state_z // 4.0)
      nd = int(state.x           // 4.0) + 48 * int(state.z           // 4.0)
      if ad != nd:
        try: self.level.char_pos[ad].remove(self)
        except (IndexError, ValueError): pass
        if 0 <= nd < 2304:
          self.level.char_pos[nd].append(self)
          
      self.last_state_x = state.x
      self.last_state_z = state.z
      
      self.speed.set_xyz(state.x - self.x, state.y - self.y, state.z - self.z)
      teleport = self.speed.length() > 2.0
      if teleport: # Guess it is a teleportation
        if self.player and self.queue.local:
          #self.level.parent.camera.zap()
          self.level.parent.camera += self.speed
          
        self.speed.set_xyz(0.0, 0.0, 0.0)
        self.set_xyz(state.x, state.y, state.z)
        self.zap() # Zap rays
        
      if state.animation and (self.current_animation != state.animation):
        if self.kill_rays: self.kill_rays = None # Ensure character's killpoint is no longer active
        
        if self.current_animation:
          self.perso.animate_clear_cycle(self.current_animation)
        self.current_animation = state.animation
        self.perso.animate_reset()
        if (self.current_animation == "combat0c") or (self.current_animation == "combat1c"):
          self.perso.animate_blend_cycle(self.current_animation, 1.0, 0.0)
        else:
          self.perso.animate_blend_cycle(self.current_animation)
          
    if state: self.reuseable_state = state
    
  def advance_time(self, proportion):
    soya.World.advance_time(self, proportion)
    
    self.add_mul_vector(proportion, self.speed)
    self.rotate_lateral(proportion * self.rotation_speed)
    
  def big_round(self):
    if self.queue.local and hasattr(self.queue.controler, "big_round"): self.queue.controler.big_round()
    
    # Check LOD
    if self.lod:
      if self.distance_to(self.parent.parent.camera) < 15.0:
        self.lod = 0
        try:
          self.perso.detach("perso_poor")
          self.perso.attach("perso")
        except: pass
    else:
      if self.distance_to(self.parent.parent.camera) > 17.0:
        self.lod = 1
        try:
          self.perso.detach("perso")
          self.perso.attach("perso_poor")
        except: pass
        
  def hurt(self, life, caster, localisation = None, vector = None, can_resist = 1):
    if self.life > 0.0:
      Blood(self.level).move(localisation or self.center)
      
      old_life = self.life
      
      if can_resist:
        resistance = self.resistance
        for item in self.items: resistance += item.resistance_bonus(caster)
        self.life -= life / resistance
      else:
        self.life -= life
        
      if   self.life <= 0.0: self.kill(caster)
      elif self.queue.local:
        if (old_life > 0.5) and (self.life <= 0.5):
          self.queue.controler.animate_and_cancel(BigHurted(self, vector))
        else:
          self.queue.controler.animate_and_cancel(Hurted(self, vector))
          
  def kill(self, caster):
    if self.queue.local:
      self.queue.controler.controlers *= 0
      self.queue.controler.animate(Die(self))
      
  def die(self):
    for item in self.items[:]: self.drop_item(item)
    if self.player and self.queue.local:
      if self.player and self.queue.local: soya.IDLER.end_game()
    else:
      self.level.remove_character(self)
      
  def display_name(self):
    return "%s (%s)" % (getattr(self, "name", ""), _(self.__class__.__name__))
  
  def kesako(self): return _("__kesako__%s" % self.__class__.__name__)
  
  def __getstate__(self):
    state = soya.World.__getstate__(self)
    if state is self.__dict__: state = state.copy()
    if state.has_key("new_center"):     del state["new_center"]
    if state.has_key("ground"):         del state["ground"]
    if state.has_key("last_ground"):    del state["last_ground"]
    if state.has_key("ground_normal"):  del state["ground_normal"]
    if state.has_key("future_speed_y"): del state["future_speed_y"]
    state["current_animation"] = None
    state["level"] = None # Do NOT send the level (too big !)
    return state
  
  def __setstate__(self, state):
    state = soya.World.__setstate__(self, state)
    self.level = None
    if self.queue.local:
      self.ground         = Point()
      self.ground_normal  = Vector()
      self.last_ground    = Point()
      self.future_speed_y = 0.0
      self.new_center     = Point(soya.IDLER.scene, self.x, self.y + self.radius_y, self.z)
  
  def set_controler(self, controler):
    if controler:
      if not self.queue.local: self.queue = LocalQueue(self)
      controler.set_character(self)
      self.queue.controler = controler
    else:
      self.queue = RemoteQueue(self)
      
      
  def __repr__(self):
    name = getattr(self, "name", "")
    if isinstance(name, unicode): name = name.encode("latin")
    if self.player:
      if self.queue.local: return "<Character %s ID %s (local player) %s,%s,%s>"  % (name, self.id, self.x, self.y, self.z)
      else:                return "<Character %s ID %s (remote player) %s,%s,%s>" % (name, self.id, self.x, self.y, self.z)
    else:
      if self.queue.local: return "<Character %s ID %s (local IA) %s,%s,%s>" % (name, self.id, self.x, self.y, self.z)
      else:                return "<Character %s ID %s (remote IA) %s,%s,%s>" % (name, self.id, self.x, self.y, self.z)
      
      
def Die(self):
  yield State(self, animation = "mort")
  for i in range(self.die_duration1): yield State(self, z = 0.04, animation = "")
  for i in range(self.die_duration2): yield State(self, animation = "")
  yield DieState(self) # => DIE
  
def Hurted(self, vector):
  if isinstance(vector, Vector):
    vector %= self
    vector.set_length(0.1)
    x = vector.x
    z = vector.z
  else:
    x = 0.0
    z = 0.1
  yield State(self, x, 0.1, z, animation = "couic")
  for i in range(6): yield State(self, x, 0.0, z, animation = "couic")
  yield State(self, x / 2.0, 0.0, z / 2.0)
  yield State(self, x / 4.0, 0.0, z / 4.0)
  
def BigHurted(self, vector):
  if isinstance(vector, Vector):
    vector %= self
    vector.set_length(0.25)
    x = vector.x
    z = vector.z
  else:
    x = 0.0
    z = 0.25
  yield State(self, x, 0.3, z, animation = "couic")
  for i in range(30): yield State(self, x, 0.0, z, animation = "couic")
  yield State(self, x / 2.0, 0.0, z / 2.0)
  yield State(self, x / 4.0, 0.0, z / 4.0)
  
  
def WeaponHurted(self):
  if   self.current_animation == "combat0": anim = "combat0c"
  elif self.current_animation == "combat1": anim = "combat1c"
  else:
    _P.clone(self.weapon)
    _P.convert_to(self)
    if _P.x < 0.0: anim = "combat0c"
    else:          anim = "combat1c"
    
  yield State(self, 0.0, 0.1, 0.0, animation = anim)
  for i in range(12):
    yield State(self, 0.0, 0.0, 0.0, animation = anim)
  
  
import random

class Blood(soya.Particles):
  def __init__(self, parent = None, material = None, nb_particles = 20):
    soya.Particles.__init__(self, parent, material or soya.PARTICLE_DEFAULT_MATERIAL, nb_particles)
    self.auto_generate_particle = 1
    self.set_colors((1.0, 0.1, 0.1, 1.0), (1.0, 0.1, 0.1, 1.0), (0.8, 0.1, 0.1, 1.0), (0.4, 0.0, 0.0, 1.0))
    self.set_sizes((0.7, 0.7))
    self.life = 11
    self.max_particles_per_round = 2
    
  def generate(self, index):
    angle = random.uniform(0.0, 6.2834)
    sx = math.cos(angle)
    sz = math.sin(angle)
    l = 0.06 / math.sqrt(sx * sx + 1.0 + sz * sz)
    self.set_particle(index, random.uniform(0.5, 1.5), sx * l, 3 * l, sz * l, 0.0, -0.01, 0.0)
    
  def begin_round(self):
    soya.Particles.begin_round(self)
    if self.life == 1:
      self.auto_generate_particle = 0
      self.removable = 1
      self.life = 0
    else: self.life -= 1

class Eclat(soya.Volume):
  def __init__(self, parent = None):
    soya.Volume.__init__(self, parent, soya.Shape.get("eclat"))
    self.solid = 0
    #soya.Material.get("eclat").additive_blending = 1
    #soya.Material.get("eclat").save()
    self.angle = 0.0
    
  def advance_time(self, proportion):
    self.angle += 0.08 * proportion
    f = math.cos(self.angle)
    if f < 0.0:
      if self.parent: self.parent.remove(self)
    else:
      self.set_scale_factors(f, f, f)


