# Balazar
# Copyright (C) 2003-2005 Jean-Baptiste LAMY
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import sys, os, os.path, math, bisect, random
import soya, soya.opengl as soyaopengl, soya.widget as widget
from soya import Point, Vector

import balazar.base       as base
import balazar.vegetation as vegetation
import balazar.building   as building
import balazar.globdef    as globdef


class Idler(soya.Idler):
  def __init__(self, scene):
    soya.Idler.__init__(self, scene)
    self.scene = scene
    
    self.min_frame_duration = globdef.MIN_FRAME_DURATION
    
  def end_game(self, result = None):
    self.stop(result)
    

def CREATE(level_name):
  level = random_level()
  level.X, level.Y = level.name = level_name
  return level

class Level(soya.World):
  """A game level.
Level is a subclass of soya.game.level.Level, which is itself a subclass of soya.soya.World."""
  def __init__(self, parent = None):
    soya.World.__init__(self, parent)
    
    self.big_round_count = 0
    
    self.next_id = -1 # Level-local ID are < 0
    
    self.id2element = {}
    self.char_pos = [[] for i in range(48**2)]
    
    self.name            = ""
    self.characters      = {}
    self.active          = 0
    
    self.X = 0
    self.Y = 0
    
  def __repr__(self): return "<Level (%s, %s)>" % (self.X, self.Y)
  
  def of_class(self, Class):
    for e in self:
      if isinstance(e, Class): return e
    for e in self.static_part:
      if isinstance(e, Class): return e
      
  def add(self, element):
    #if not element.name:
    #  element.name = str(self.random.random()) [:10]
    if not hasattr(element, "id") or element.id == 0:
      element.id = self.next_id
      self.next_id -= 1
    self.id2element[element.id] = element
    soya.World.add(self, element)
    
    if isinstance(element, base.Entity):
      d = int(element.x // 4.0) + 48 * int(element.z // 4.0)
      if 0 <= d < 2304: self.char_pos[d].append(element)
  append = add
  
  def remove(self, element):
    del self.id2element[element.id]
    soya.World.remove(self, element)
    
    if isinstance(element, base.Entity):
      d = int(element.x // 4.0) + 48 * int(element.z // 4.0)
      try: self.char_pos[d].remove(element)
      except: pass
      
  def add_character(self, character):
    self.add(character)
    self.characters[character.id] = character
    
  def remove_character(self, character):
    del self.characters[character.id]
    self.remove(character)
    
  def build_hash_quad(self):
    for child in self.children:
      if isinstance(child, base.Entity):
        d = int(child.x // 4.0) + 48 * int(child.z // 4.0)
        if 0 <= d < 2304: self.char_pos[d].append(child)
        
  def set_active(self, active):
    # If the level is active, it is modified => it is no longer the same than the one
    # saved in the corresponding file.
    if active: self.filename = ""
    
    self.active = active
    
    if active:
      if not hasattr(self, "random"): self.random = random.Random()
      
      
  def begin_round(self):
    soya.World.begin_round(self)
    
    if self.big_round_count == 0:
      self.big_round_count = 15
      self.big_round()
      
    else: self.big_round_count -= 1
    
  def advance_time(self, proportion):
    soya.World.advance_time(self, proportion)
    self.atmosphere.advance_time(proportion)
    
  def big_round(self):
    self.atmosphere.big_round()
    
    for child in self.children:
      if hasattr(child, "big_round"): child.big_round()
      
#   def __getstate__(self):
#     state = super(Level, self).__getstate__()
#     del state["char_pos"]
#     del state["need_send_state"]
#     #print state
#     return state
  
#   def __setstate__(self, state):
#     ng_level.Level.__setstate__(self, state)
#     self.char_pos = [[] for i in range(48**2)]
#     self.build_hash_quad()
    
    
WEATHER = ["sunny", "rain", "snow"]

DAY_NIGHT_DATA = {
  "sunny" : [
  (4.0, # Time
   0.0, 70.0, # Fog distance
   0.0, 0.0, 0.0, # Fog color
   0.0, 0.3, 1.0, # Sky color
   0.15, 0.15, 0.3, # Ambient
   0.3, 0.3, 0.5, # Sun diffuse color
   ),
  (8.0, # Time
   20.0, 80.0, # Fog distance
   0.4, 0.45, 0.7, # Fog color
   1.5, 1.0, 0.8, # Sky color
   0.3, 0.3, 0.4, # Ambient
   1.0, 0.8, 0.4, # Sun diffuse color
   ),
  (12.0, # Time
   50.0, 100.0, # Fog distance
   0.2, 0.5, 0.7, # Fog color
   2.0, 2.0, 2.0, # Sky color
   0.4, 0.4, 0.4, # Ambient
   0.9, 0.9, 0.7, # Sun diffuse color
   ),
  (16.0, # Time
   40.0, 95.0, # Fog distance
   0.45, 0.6, 1.0, # Fog color
   0.0, 0.0, 1.0, # Sky color
   0.3, 0.3, 0.4, # Ambient
   0.7, 0.7, 0.5, # Sun diffuse color
   ),
   (19.0, # Time
   30.0, 80.0, # Fog distance
   0.3, 0.0, 0.6, # Fog color
   1.5, 1.2, 0.2, # Sky color
   0.2, 0.2, 0.6, # Ambient
   0.7, 0.7, 0.5, # Sun diffuse color
   ),
   (24.0, # Time -- This one is MANDATORY !!!
   5.0, 70.0, # Fog distance
   0.0, 0.0, 0.0, # Fog color
   0.2, 0.0, 1.0, # Sky color
   0.15, 0.15, 0.4, # Ambient
   0.3, 0.3, 0.5, # Sun diffuse color
   ),
 ],
  "rain" : [
  (4.0, # Time
   0.0, 65.0, # Fog distance
   0.0, 0.0, 0.0, # Fog color
   0.2, 0.3, 0.5, # Sky color
   0.15, 0.15, 0.2, # Ambient
   0.2, 0.2, 0.3, # Sun diffuse color
   ),
  (8.0, # Time
   5.0, 70.0, # Fog distance
   0.4, 0.45, 0.5, # Fog color
   0.2, 0.2, 0.3, # Sky color
   0.3, 0.3, 0.3, # Ambient
   0.6, 0.4, 0.3, # Sun diffuse color
   ),
  (12.0, # Time
   25.0, 95.0, # Fog distance
   0.4, 0.5, 0.6, # Fog color
   0.3, 0.3, 0.3, # Sky color
   0.3, 0.3, 0.3, # Ambient
   0.5, 0.5, 0.5, # Sun diffuse color
   ),
  (16.0, # Time
   25.0, 90.0, # Fog distance
   0.4, 0.4, 0.6, # Fog color
   0.2, 0.2, 0.4, # Sky color
   0.3, 0.3, 0.3, # Ambient
   0.4, 0.4, 0.4, # Sun diffuse color
   ),
   (19.0, # Time
   15.0, 80.0, # Fog distance
   0.4, 0.3, 0.5, # Fog color
   0.4, 0.3, 0.2, # Sky color
   0.2, 0.2, 0.3, # Ambient
   0.3, 0.3, 0.2, # Sun diffuse color
   ),
   (24.0, # Time -- This one is MANDATORY !!!
   0.0, 65.0, # Fog distance
   0.0, 0.0, 0.0, # Fog color
   0.2, 0.2, 0.8, # Sky color
   0.15, 0.15, 0.3, # Ambient
   0.2, 0.2, 0.3, # Sun diffuse color
   ),
  ],
  "snow" : [
  (4.0, # Time
   0.0, 70.0, # Fog distance
   0.0, 0.0, 0.0, # Fog color
   0.6, 0.0, 0.4, # Sky color
   0.15, 0.15, 0.3, # Ambient
   0.3, 0.2, 0.3, # Sun diffuse color
   ),
  (8.0, # Time
   0.0, 80.0, # Fog distance
   0.6, 0.5, 0.7, # Fog color
   1.5, 1.0, 0.8, # Sky color
   0.3, 0.3, 0.4, # Ambient
   0.5, 0.4, 0.3, # Sun diffuse color
   ),
  (12.0, # Time
   5.0, 100.0, # Fog distance
   0.5, 0.7, 0.7, # Fog color
   2.0, 2.0, 2.0, # Sky color
   0.4, 0.4, 0.4, # Ambient
   0.6, 0.6, 0.4, # Sun diffuse color
   ),
  (16.0, # Time
   3.0, 95.0, # Fog distance
   0.5, 0.8, 0.8, # Fog color
   0.0, 0.0, 1.0, # Sky color
   0.3, 0.3, 0.4, # Ambient
   0.4, 0.4, 0.2, # Sun diffuse color
   ),
   (19.0, # Time
   0.0, 80.0, # Fog distance
   0.3, 0.2, 0.6, # Fog color
   1.5, 1.2, 0.7, # Sky color
   0.2, 0.2, 0.6, # Ambient
   0.4, 0.4, 0.2, # Sun diffuse color
   ),
   (24.0, # Time -- This one is MANDATORY !!!
   0.0, 70.0, # Fog distance
   0.0, 0.0, 0.0, # Fog color
   0.4, 0.0, 0.3, # Sky color
   0.15, 0.15, 0.4, # Ambient
   0.3, 0.2, 0.3, # Sun diffuse color
   ),
  ],
  }
WEATHER_DURATION = {
  "sunny" : 24.0,
  "rain"  : 12.0,
  "snow"  : 12.0,
  }

class SunSprite(soya.Sprite, base.PhotoScoreable):
  def __init__(self, parent = None):
    soya.Sprite.__init__(self, parent, soya.Material.get("sun"))
    self.color = (1.0, 0.7, 0.5, 1.0)
    self.width = self.height = 20.0
    
  def get_photo_score(self, photo, camera):
    return 1000.0, 0.0, 0.1
  
  def display_name(self):
    return _("sun")
  
class StarSprite(soya.Sprite, base.PhotoScoreable):
  def __init__(self, parent = None):
    soya.Sprite.__init__(self, parent, soya.Material.get("star"))
    
  def get_photo_score(self, photo, camera):
    return 100.0, 0.0, 0.1
  
  def display_name(self):
    return _("star")
  
  
class DayNightCycle(soya.SkyAtmosphere):
  def __init__(self, parent = None):
    soya.SkyAtmosphere.__init__(self)
    self.fog      = 1
    self.fog_type = 0
    self.skyplane = 1
    
    self.time    = 7.001 # In hour
    self.weather = "sunny"
    self.weather_duration = 0.1
    self.old_data = DAY_NIGHT_DATA["sunny"][0]
    self.old_right = 1
    self.sun_pos = 1.0
    self.day     = 1
    self.rain    = self.snow = None
    
    self.delta   = 0.03
    
    if parent:
      self.sun = soya.Light(parent)
      self.sun.directional = 1
      self.sun.rotate_vertical(7.001 / 24.0 * 360.0 + 90.0)
      
      self.astral_world = soya.World(parent)
      self.astral_world.atmosphere = soya.SkyAtmosphere()
      self.astral_world.atmosphere.ambient = (1.0, 1.0, 1.0, 1.0)
      self.astral_world.rotate_vertical(7.001 / 24.0 * 360.0 + 90.0)
      
      self.sun_sprite = SunSprite(self.astral_world)
      self.sun_sprite.z = 90.0
      
      for i in range(40):
        star = StarSprite(self.astral_world)
        star.color = (random.uniform(0.0, 1.0), random.uniform(0.0, 1.0), random.uniform(0.0, 1.0), 1.0)
        star.size = star.width = star.height = random.uniform(0.5, 2.0)
        
        star.x = random.uniform(-90.0,  90.0)
        star.y = random.uniform(-90.0,  90.0)
        star.z = random.uniform(-90.0, -20.0)
        l = self.astral_world.distance_to(star) / 99.0
        star.x /= l
        star.y /= l
        star.z /= l
        
      parent.atmosphere = self
      
  def get_photo_score(self, photo, camera):
    if self.time <  5.0: return  0.0, 12.0
    if self.time <  6.0: return  0.0,  2.0
    if self.time <  9.0: return  5.0,  0.0
    if self.time < 15.0: return  1.0,  0.0
    if self.time < 17.0: return  3.0,  0.0
    if self.time < 20.5: return 11.0,  3.0
    return                       0.0, 10.0
  
  def big_round(self, delta = 0.03):
    self.delta = delta
    
    #if not self.rain:
    #  self.rain = Rain(self)
    #  soya.root_widget.insert(1, self.rain) # 0 is camera
    #if not self.snow:
    #  self.snow = Snow()
    #  soya.root_widget.insert(1, self.snow) # 0 is camera
      
    self.time += delta
    if self.time >= 24.0:
      self.time -= 24.0
      self.day  += 1
      
    data = DAY_NIGHT_DATA[self.weather]
    right = bisect.bisect(data, (self.time,))
    
    self.weather_duration -= delta
    
    if right != self.old_right: # Can change weather here
      self.old_right = right
      self.old_data  = data[right - 1]
      
      if (self.weather == "rain") and not self.rain:
        self.astral_world.visible = 0
        self.rain = Rain(self)
        soya.root_widget.insert(1, self.rain) # 0 is camera
        
      if self.weather_duration < 0.0:
        if   self.weather == "rain":
          self.astral_world.visible = 1
          soya.root_widget.remove(self.rain)
          self.rain = None
        elif self.weather == "snow":
          soya.root_widget.remove(self.snow)
          self.snow = None
          
        self.weather = self.astral_world.parent.random.choice(WEATHER)
        self.weather_duration = random.uniform(4.0, WEATHER_DURATION[self.weather])
        data = DAY_NIGHT_DATA[self.weather]
        
        if self.weather == "snow":
          self.snow = Snow()
          soya.root_widget.insert(1, self.snow) # 0 is camera

    d1 = self.old_data
    d2 = data[right]
    
    if d1[0] < d2[0]: f2 = (self.time - d1[0]) / (d2[0] - d1[0])
    else:             f2 = (24.0 + self.time - d1[0]) / (24.0 + d2[0] - d1[0])
    f1 = 1.0 - f2
    
    self.fog_start   = f1 * d1[1] + f2 * d2[1]
    self.fog_end     = (f1 * d1[2] + f2 * d2[2])
    self.fog_color   = self.bg_color = (f1 * d1[ 3] + f2 * d2[ 3], f1 * d1[ 4] + f2 * d2[ 4], f1 * d1[ 5] + f2 * d2[ 5], 1.0)
    self.sky_color   =                 (f1 * d1[ 6] + f2 * d2[ 6], f1 * d1[ 7] + f2 * d2[ 7], f1 * d1[ 8] + f2 * d2[ 8], 1.0)
    self.ambient     =                 (f1 * d1[ 9] + f2 * d2[ 9], f1 * d1[10] + f2 * d2[10], f1 * d1[11] + f2 * d2[11], 1.0)
    self.sun.diffuse =                 (f1 * d1[12] + f2 * d2[12], f1 * d1[13] + f2 * d2[13], f1 * d1[14] + f2 * d2[14], 1.0)
    
    if   (self.sun_pos == 1.0) and (self.time > 18.0): # the sun becomes the moon and travel backward
      self.sun_pos = -1.0
    elif (self.sun_pos == -1.0) and (6.0 < self.time < 18.0):
      self.sun_pos = 1.0
    
  def advance_time(self, proportion):
    a = self.delta * proportion * 0.9375 #360.0 / 16.0 / 24.0 
    self.sun.rotate_vertical(self.sun_pos * a)
    self.astral_world.rotate_vertical(      a)
    
    self.astral_world.move(self.astral_world.parent.parent.camera)
    
  def __setstate__(self, state):
    soya.SkyAtmosphere.__setstate__(self, state)
    if self.rain:
      self.rain.__init__(self)
      soya.root_widget.insert(1, self.rain) # 0 is camera
    if self.snow:
      self.snow.__init__()
      soya.root_widget.insert(1, self.snow) # 0 is camera
    
    
class Rain(widget.Widget):
  def __init__(self, daynight, parent = None):
    self.w = 640
    self.h = 480
    self.next_thunder = 10
    self.daynight = daynight
    self.particles = [[random.uniform(0.0, self.w), random.uniform(-20.0, self.h), random.uniform(15.0, 50.0), random.uniform(0.5, 1.0)] for i in range(30)]
    
    widget.Widget.__init__(self, parent)
    
  def __getstate__(self): return None
  def __setstate__(self, state): pass
    
  def resize(self, parent_left, parent_top, parent_width, parent_height):
    w = soya.get_screen_width()
    h = soya.get_screen_height()
    fw = float(w) / self.w
    fh = float(h) / self.h
    self.w = w
    self.h = h
    
    for p in self.particles:
      p[0] *= fw
      p[1] *= fh
      p[2] *= fh
      
  def render(self):
    f = soya.IDLER.scene.transform_vector(1.0, 0.0, 0.0, soya.IDLER.camera)[0] / 2.0
    
    soya.DEFAULT_MATERIAL.activate()
    soyaopengl.glLineWidth(2.0)
    soyaopengl.glBegin(soyaopengl.GL_TRIANGLES)
    for p in self.particles:
      if p[1] > 0.6 * self.h + p[2] * 2.5: p[1] = 0.0
      if   p[0] > self.w: p[0] -= self.w
      elif p[0] < 0.0:    p[0] += self.w
      
      soyaopengl.glColor4f(p[3], 0.8, 1.0, 1.0)
      soyaopengl.glVertex2f(p[0], p[1])
      p[0] += p[2] *  f
      p[1] += p[2]
      soyaopengl.glVertex2f(p[0], p[1])
      soyaopengl.glVertex2f(p[0] + 3, p[1])
      p[0] -= p[2] * 0.5 * f
      p[1] -= p[2] * 0.5
      
    soyaopengl.glEnd()
    
    self.next_thunder -= 1
    if   self.next_thunder == 0:
      self.materials = soya.Material._alls.values()
      #for material in self.materials:
      #  material.old_data = material.separate_specular, material.specular, material.shininess
      #  material.separate_specular = 1
      #  material.specular = (1.0, 1.0, 1.0, 1.0)
      #  material.shininess = 0.0
      
      self.daynight.fog_color = (1.0, 1.0, 1.0, 1.0)
      self.daynight.bg_color  = self.daynight.sky_color  = (1.0, 1.0, 1.0, 1.0)
      self.daynight.fog_start = -20.0
    #elif -20 < self.next_thunder < -5:
    #  f = float(-self.next_thunder - 5) / 15.0 * 128.0
    #  for material in self.materials:
    #    material.shininess = f
        
    #elif self.next_thunder == -20:
    #  for material in self.materials:
    #    material.separate_specular, material.specular, material.shininess = material.old_data
        
      self.next_thunder = random.randint(50, 200)
      
    
    
class Snow(widget.Widget):
  def __init__(self, parent = None):
    widget.Widget.__init__(self, parent)
    self.w = 640
    self.h = 480
    
    self.particles = [[random.uniform(0.0, self.w), 0.0, random.uniform(20.0, 50.0), random.uniform(0.5, 1.0)] for i in range(15)]
    self.old_cam_x = 0.0
    self.old_cam_y = 0.0
    
    self.material = soya.Material.get("star")
    
  def __getstate__(self): return None
  def __setstate__(self, state): pass
  
  def resize(self, parent_left, parent_top, parent_width, parent_height):
    w = soya.get_screen_width()
    h = soya.get_screen_height()
    fw = float(w) / self.w
    fh = float(h) / self.h
    self.w = w
    self.h = h
    
    for p in self.particles:
      p[0] *= fw
      p[1] *= fh
      
    self.v1 = Vector(soya.IDLER.scene, 0.0, 0.0, -1.0)
    self.v2 = Vector(soya.IDLER.camera, 0.0, 0.0, -1.0)
    
  def render(self):
    if not hasattr(self, "v1"): return
    
    a = self.v1.angle_to(self.v2)
    if self.v1.cross_product(self.v2).y < 0.0: delta_x = (self.old_cam_x - a) * self.w * 0.5
    else:                                      delta_x = (a - self.old_cam_x) * self.w * 0.5
    self.old_cam_x = a
    
    delta_y = (soya.IDLER.camera.y - self.old_cam_y) * self.h * 0.1
    self.old_cam_y = soya.IDLER.camera.y
    
    self.material.activate()
    soyaopengl.glEnable(soyaopengl.GL_BLEND)
    soyaopengl.glPointSize(20.0)
    soyaopengl.glBegin(soyaopengl.GL_QUADS)
    for p in self.particles:
      if   p[1] > self.h - 35: p[1] -= self.h
      elif p[1] < - 35: p[1] += self.h
      if   p[0] < 0.0: p[0] += self.w
      elif p[0] > self.w: p[0] -= self.w
      
      soyaopengl.glColor4f(0.5, 0.3, 1.0, 1.0)
      soyaopengl.glTexCoord2f(0.0, 0.0); soyaopengl.glVertex2f(p[0], p[1])
      soyaopengl.glTexCoord2f(1.0, 0.0); soyaopengl.glVertex2f(p[0], p[1] + p[2])
      soyaopengl.glTexCoord2f(1.0, 1.0); soyaopengl.glVertex2f(p[0] + p[2], p[1] + p[2])
      soyaopengl.glTexCoord2f(0.0, 1.0); soyaopengl.glVertex2f(p[0] + p[2], p[1])
      p[0] += delta_x / p[2]
      p[1] += p[2] * (0.1 + 0.02 * delta_y)
      
    soyaopengl.glEnd()
    soya.DEFAULT_MATERIAL.activate()
    
      
    









  

LAND_SIZE = 129
LAND_DIM  = 1.5 * LAND_SIZE

def create_level():
  level = Level()
  level.random = random.Random()
  level.static_part = soya.World(level)
  level.buildings          = []
  level.interesting_places = []
  level.down               = Vector(level, 0.0, -1.0, 0.0)
  return level
  
def end_level(level):
  level.build_hash_quad()
  del level.buildings
  del level.interesting_places
  del level.down
  
  
def create_land(level):
  land = soya.Land(None, LAND_SIZE)
  land.texture_factor = 0.25
  return land

def random_land(level, land):
  # Default height
  for i in range(LAND_SIZE):
    for j in range(LAND_SIZE):
      land.set_height(i, j, random.uniform(0.18, 0.23))
      
  for n in range(int(random.uniform(0.001, 0.0015) * LAND_SIZE * LAND_SIZE)):
    t = random.random()
    x = random.randrange(LAND_SIZE)
    y = random.randrange(LAND_SIZE)
    if   t < 0.1:
      # Pic
      r = random.randrange(5, 20)
      h = random.uniform(0.5, 2.5)
      
      for i in range(max(0, x - r), min(LAND_SIZE - 1, x + r)):
        for j in range(max(0, y - r), min(LAND_SIZE - 1, y + r)):
          d = (r - math.sqrt((i - x) * (i - x) + (j - y) * (j - y))) / float(r)
          if d > 0.0:
            d = d * d
            land.set_height(i, j, min(1.0, d * h + land.get_height(i, j)))
            
    elif t < 0.35:
      # Plateau
      r1 = random.randrange(10, 20)
      r2 = random.randrange(r1 + 5, 40)
      h  = random.uniform(0.05, 0.3)
      
      for i in range(max(0, x - r2), min(LAND_SIZE - 1, x + r2)):
        for j in range(max(0, y - r2), min(LAND_SIZE - 1, y + r2)):
          d = (r2 - math.sqrt((i - x) * (i - x) + (j - y) * (j - y))) / float(r2 - r1)
          if d > 1.0: d = 1.0
          if d > 0.0:
            d = 0.5 - (math.cos(d * math.pi) / 2.0)
            land.set_height(i, j, min(1.0, d * h + land.get_height(i, j)))
            
    elif t < 0.6:
      # Hill
      r = random.randrange(5, 35)
      h = random.uniform(0.1, 0.4)
      
      for i in range(max(0, x - r), min(LAND_SIZE - 1, x + r)):
        for j in range(max(0, y - r), min(LAND_SIZE - 1, y + r)):
          d = (r - math.sqrt((i - x) * (i - x) + (j - y) * (j - y))) / float(r)
          if d > 0.0:
            d = 0.5 - (math.cos(d * math.pi) / 2.0)
            land.set_height(i, j, min(1.0, d * h + land.get_height(i, j)))
            
    elif t < 0.7:
      # Cratre trange...!
      r = random.randrange(15, 20)
      h = random.uniform(0.1, 0.1)
      level.interesting_places.append((x * land.scale_factor, y * land.scale_factor))
      
      for i in range(max(0, x - r), min(LAND_SIZE - 1, x + r)):
        for j in range(max(0, y - r), min(LAND_SIZE - 1, y + r)):
          d = (r - math.sqrt((i - x) * (i - x) + (j - y) * (j - y))) / float(r)
          if d > 0.0:
            d = (math.sin(d * math.pi) / 2.0) + 0.5
            land.set_height(i, j, min(1.0, d * h + land.get_height(i, j)))
            
    elif t < 0.8:
      # Cratre trange 2...!
      r = random.randrange(15, 20)
      h = random.uniform(0.1, 0.1)
      level.interesting_places.append((x * land.scale_factor, y * land.scale_factor))
      
      for i in range(max(0, x - r), min(LAND_SIZE - 1, x + r)):
        for j in range(max(0, y - r), min(LAND_SIZE - 1, y + r)):
          d = (r - math.sqrt((i - x) * (i - x) + (j - y) * (j - y))) / float(r)
          if d > 0.0:
            d = (math.cos(d * math.pi) / 2.0) + 0.5
            land.set_height(i, j, min(1.0, d * h + land.get_height(i, j)))
            
    elif t < 0.9:
      # Monticule
      r = random.randrange(5, 20)
      h = random.uniform(0.5, 1.0)
      level.interesting_places.append((x * land.scale_factor, y * land.scale_factor))
      
      for i in range(max(0, x - r), min(LAND_SIZE - 1, x + r)):
        for j in range(max(0, y - r), min(LAND_SIZE - 1, y + r)):
          d = (r - math.sqrt((i - x) * (i - x) + (j - y) * (j - y))) / float(r)
          if d > 0.0:
            d = math.sqrt(d)
            land.set_height(i, j, min(1.0, d * h + land.get_height(i, j)))
            
    else:
      # Termitire conique XXX une texture spciale ?
      r = random.randrange(5, 15)
      h = random.uniform(0.1, 0.5)
      
      for i in range(max(0, x - r), min(LAND_SIZE - 1, x + r)):
        for j in range(max(0, y - r), min(LAND_SIZE - 1, y + r)):
          d = (r - math.sqrt((i - x) * (i - x) + (j - y) * (j - y))) / float(r)
          if d > 0.0:
            land.set_height(i, j, min(1.0, d * h + land.get_height(i, j)))
            
def smooth_land(level, land, nb = 2):
  for k in range(nb):
    smoothed = [0.0] * (LAND_SIZE * LAND_SIZE)
    for i in range(1, LAND_SIZE - 1):
      for j in range(1, LAND_SIZE - 1):
        smoothed[i + LAND_SIZE * j] = (land.get_height(i, j) + land.get_height(i - 1, j) + land.get_height(i, j - 1) + land.get_height(i + 1, j) + land.get_height(i, j + 1)) / 5.0
    for i in range(1, LAND_SIZE - 1):
      for j in range(1, LAND_SIZE - 1):
        land.set_height(i, j, smoothed[i + LAND_SIZE * j])
        
def random_land_border(level, land):
  out_of_map_vertex_options = (1, 1)
  
  fall_length  = 2
  afall_length = 0
  for i in range(LAND_SIZE):
    for j in range(fall_length):
      land.set_vertex_option(i, j, *out_of_map_vertex_options)
      #land.set_vertex_color(i, j, (0.0, 0.0, 0.0, 0.0))
      land.set_height(i, j, 0.0)
    land.set_height(i, j + 1, 0.92 * land.get_height(i, j + 1))
    if i % 2:
      fall_length = max(1, fall_length + afall_length)
      if (fall_length == 1) and (afall_length == -1): afall_length = 0
      else: afall_length = max(-1, min(1, afall_length + random.randint(-1, 1)))
      
  fall_length  = 2
  afall_length = 0
  for i in range(LAND_SIZE):
    for j in range(LAND_SIZE - 1, LAND_SIZE - fall_length - 1, -1):
      land.set_vertex_option(i, j, *out_of_map_vertex_options)
      #land.set_vertex_color(i, j, (0.0, 0.0, 0.0, 0.0))
      land.set_height(i, j, 0.0)
    land.set_height(i, j - 1, 0.92 * land.get_height(i, j - 1))
    if i % 2:
      fall_length = max(1, fall_length + afall_length)
      if (fall_length == 1) and (afall_length == -1): afall_length = 0
      else: afall_length = max(-1, min(1, afall_length + random.randint(-1, 1)))

  fall_length  = 2
  afall_length = 0
  for j in range(LAND_SIZE):
    for i in range(fall_length):
      land.set_vertex_option(i, j, *out_of_map_vertex_options)
      #land.set_vertex_color(i, j, (0.0, 0.0, 0.0, 0.0))
      land.set_height(i, j, 0.0)
    land.set_height(i + 1, j, 0.92 * land.get_height(i + 1, j))
    if j % 2:
      fall_length = max(1, fall_length + afall_length)
      if (fall_length == 1) and (afall_length == -1): afall_length = 0
      else: afall_length = max(-1, min(1, afall_length + random.randint(-1, 1)))
      
  fall_length  = 2
  afall_length = 0
  for j in range(LAND_SIZE):
    for i in range(LAND_SIZE - 1, LAND_SIZE - fall_length - 1, -1):
      land.set_vertex_option(i, j, *out_of_map_vertex_options)
      #land.set_vertex_color(i, j, (0.0, 0.0, 0.0, 0.0))
      land.set_height(i, j, 0.0)
    land.set_height(i - 1, j, 0.92 * land.get_height(i - 1, j))
    if j % 2:
      fall_length = max(1, fall_length + afall_length)
      if (fall_length == 1) and (afall_length == -1): afall_length = 0
      else: afall_length = max(-1, min(1, afall_length + random.randint(-1, 1)))


def end_land(level, land):
  land.multiply_height(50.0)
  land.scale_factor = 1.5
  land.map_size = 8
  land.split_factor = 4.0
  
  level.static_part.add(land)


def check_distance(level, land, house):
  for building in level.buildings:
    if math.sqrt((house.x - building.x) ** 2 + (house.z - building.z) ** 2) < (getattr(house, "radius", 1.0) + getattr(building, "radius", 1.0)) * 1.8:
      return 0
  level.buildings.append(house)
  return 1

def put_on_ground(level, land, o, orientate = 0):
  old_solid = o.solid
  o.solid = 0
  
  r = level.raypick(o, Vector(level, 0.0, -1.0, 0.0), -1.0, 1, 1)
  
  o.solid = old_solid
  if r:
    if r[0].y <= 0.0: return 0
    o.y = (r[0] % o.parent).y
    if orientate: o.look_at_y(r[1])
    return r
  
def terraform(level, land, terraformer):
  x = int(round(terraformer.x / land.scale_factor))
  y = int(round(terraformer.z / land.scale_factor))
  r1 = terraformer.terraform_radius
  r2 = r1 * 2.0
  h  = terraformer.y
  
  for i in range(max(0, int(x - r2)), min(LAND_SIZE - 1, int(x + r2))):
    for j in range(max(0, int(y - r2)), min(LAND_SIZE - 1, int(y + r2))):
      if not land.get_vertex_option(i, j)[0]:
        d = (r2 - math.sqrt((i - x) * (i - x) + (j - y) * (j - y))) / float(r2 - r1)
        if   d > 1.0: land.set_height(i, j, h)
        elif d > 0.0: land.set_height(i, j, d * h + (1.0 - d) * land.get_height(i, j))
          
          
def place(level, land, obj, orientate = 0, uppon = None, arround = None, arround_radius = 5.0, min_x = 0.0, max_x = 10000.0, min_z = 0.0, max_z = 10000.0):
  obj.rotate_lateral(random.uniform(0.0, 360.0))
  radius = getattr(obj, "radius", 3.0) * 3.0
  if   uppon:
    for i in range(100):
      obj.set_xyz(uppon.x + random.uniform(-uppon.radius, uppon.radius), 1000.0, uppon.z + random.uniform(-uppon.radius, uppon.radius))
      impact = put_on_ground(level, land, obj, orientate)
      if impact and impact[0].parent.is_inside(uppon): break
      
  elif arround:
    arround = arround % level
    for i in range(100):
      obj.set_xyz(arround.x + random.uniform(-arround_radius, arround_radius), 1000.0, arround.z + random.uniform(-arround_radius, arround_radius))
      if put_on_ground(level, land, obj, orientate): break
      
  else:
    for i in range(100):
      obj.set_xyz(
        min(max_x, max(min_x, random.uniform(radius, LAND_DIM - radius))),
        1000.0,
        min(max_z, max(min_z, random.uniform(radius, LAND_DIM - radius))),
        )
      if put_on_ground(level, land, obj, orientate) and check_distance(level, land, obj): break
  if isinstance(obj, base.Terraformer): terraform(level, land, obj)
  if hasattr(obj, "placed"): obj.placed()

def place_character(level, land, character, pos):
  character.rotate_lateral(random.uniform(0.0, 360.0))
  radius = getattr(character, "radius", 3.0) * 3.0
  p = pos % level
  character.move(p)
  p.y = 1000.0
  r = level.raypick(p, level.down, -1.0, 1, 1)
  if not r:
    for i in range(100):
      character.set_xyz(random.uniform(radius, LAND_DIM - radius), 1000.0, random.uniform(radius, LAND_DIM - radius))
      r = level.raypick(p, level.down, -1.0, 1, 1)
      if r: break
    else: return
    
  character.y = r[0].y
  character.set_level(level)
  if hasattr(character, "placed"): character.placed()
  
  
def add_flower(level, land, nb = 40):
  grass2 = soya.Material.get("grass3")
  for i in range(nb):
    x = random.randrange(LAND_SIZE)
    y = random.randrange(LAND_SIZE)
    r = random.randint(1, 3)
    if r == 3: level.interesting_places.append((x * land.scale_factor, y * land.scale_factor))
    
    for i in range(max(0, x - r), min(LAND_SIZE - 1, x + r)):
      for j in range(max(0, y - r), min(LAND_SIZE - 1, y + r)):
        d = (r - math.sqrt((i - x) * (i - x) + (j - y) * (j - y))) / float(r)
        if (d >= 0.0) and land.get_height(i, j) < 30:
          land.set_material(i, j, grass2)
  

def add_random_item(level, land):
  import balazar.item as balazar_item
  item = random.choice(balazar_item.SMALL_TREASURE)(level)
  for i in range(100):
    if level.interesting_places and random.random() < 0.5:
      pos = level.interesting_places.pop()
      item.set_xyz(pos[0], 1000.0, pos[1])
    else: item.set_xyz(random.uniform(10.0, LAND_DIM - 20.0), 1000.0, random.uniform(10.0, LAND_DIM - 20.0))
    if put_on_ground(level, land, item, 1): break
  if item.can_rotate: item.turn_incline(90.0)
  
  

  
def random_level():
  from balazar.base       import DecorativeEntity
  from balazar.vegetation import UP
  import balazar.npc as npc
  
  level = create_level()
  DayNightCycle(level)
  
  land = create_land(level)
  random_land       (level, land)
  
  # Add a pic, to hide the TrueTemple
  r = 15
  h = 2.5
  x = 83
  y = 83
  for i in range(max(0, x - r), min(LAND_SIZE - 1, x + r)):
    for j in range(max(0, y - r), min(LAND_SIZE - 1, y + r)):
      d = (r - math.sqrt((i - x) * (i - x) + (j - y) * (j - y))) / float(r)
      if d > 0.0:
        d = d * d
        land.set_height(i, j, min(1.0, d * h + land.get_height(i, j)))
        
  smooth_land       (level, land)
  
  land.set_material_layer(soya.Material.get("grass2"), 0.0 , 0.24)
  land.set_material_layer(soya.Material.get("grass" ), 0.24, 0.54)
  land.set_material_layer(soya.Material.get("snow"  ), 0.54, 0.6 )
  land.set_material_layer(soya.Material.get("ice"   ), 0.6 , 1.0 )
  
  random_land_border(level, land)
  end_land          (level, land)
  
  from balazar.discussion import Discussion, DiscussionScript
  
  def temple_seen():
    from balazar.player import CURRENT_CHARACTER
    CURRENT_CHARACTER.set_spc("know_temple")
    level.remove(temple_discu1)
    
    def item_listener(item):
      print item, item.score, item.clazz_content
      from balazar.item     import Photo
      from balazar.vehicule import FlyingBoat
      if isinstance(item, Photo) and (item.score >= 300) and (building.TrueTemple in item.clazz_content):
        for morkul in level.characters.values():
          if isinstance(morkul, npc.Morkul): morkul.change_relation(CURRENT_CHARACTER, -1.0)
        [e for e in level if isinstance(e, FlyingBoat)][0].speed.set_xyz(0.0, 0.2, 0.0)
        Discussion(CURRENT_CHARACTER, u"", u"""J'ai la photo du temple ! Nous avons russi ! \
Je dois maintenant rejoindre mon Drivoile... Attention ! Je crois que les Morkuls m'en veulent !""").activate()
        
    CURRENT_CHARACTER.item_listeners.append(item_listener)

  # lambda : temple.activated
  def test():
    print temple, temple.activated
    return temple.activated
  
  temple = building.TrueTemple(level.static_part)
  place(level, land, temple, min_x = 150.0, min_z = 150.0)
  temple_discu1 = DiscussionScript(level, Discussion(None, u"", _(u"Voil le temple que je cherchais... Mais \
il n'a pas t entretenu depuis des annes ! Il est couvert de champignon. Il faut que \
je le nettoie suffisament pour prendre une photo  plus de 300 points."),
    condition = lambda : temple.activated,
    choices = [
      Discussion(None, _(u"Au boulot !"), u"", action = temple_seen),
    ]), 25.0)
  temple_discu1.move(temple)
  
  
  morkul_temple = building.MorkulTemple(level.static_part)
  place(level, land, morkul_temple, max_x = 100.0, max_z = 100.0)
  def morkul_temple_seen():
    from balazar.player import CURRENT_CHARACTER
    CURRENT_CHARACTER.set_spc("know_morkul_temple_is_bad")
    level.remove(morkul_temple_discu1)
    
  morkul_temple_discu1 = DiscussionScript(level, Discussion(None, u"", _(u"Cette chose serait le temple \
que je cherche ? Non c'est impossible ! Il doit y avoir une erreur !"),
    choices = [
      Discussion(None, _(u"Je vais chercher le vrai temple cette fois-ci !"), u"", action = morkul_temple_seen),
    ]), 10.0)
  morkul_temple_discu1.move(morkul_temple)

  place(level, land, building.Market(level.static_part))
  for i in range( 3): place(level, land, building.Farm (level.static_part), max_x = 130.0, max_z = 130.0)
  for i in range( 3): place(level, land, building.Tower(level.static_part))
  houses = [house for house in level.buildings if isinstance(house, building.Farm)]
  
  land.reinit() # XXX in end_land ???
  
  import balazar.item
  axe = balazar.item.Axe(level)
  axe.blessed = 0 # Avoid that Morkuls take it
  p = soya.Point(houses[1], 15.0, 0.0, 15.0)
  place(level, land, axe, arround = p, arround_radius = 15.0)
  
  for i in range(10):
    d = DecorativeEntity(level.static_part, soya.Shape.get("dent"))
    place(level, land, d, 1)
    s = random.uniform(0.5, 2)
    d.scale(s, s, s)
  for i in range(4):
    place(level, land, vegetation.Mushroom  (level), uppon   = temple)
  for i in range(2):
    place(level, land, vegetation.Mushroom  (level, grow = 0.0), arround = axe, arround_radius = 1.3)
  for i in range(2):
    place(level, land, vegetation.Mushroom  (level), arround = axe, arround_radius = 1.3)
  for i in range(8): place(level, land, vegetation.Mushroom  (level))
  for i in range(4): place(level, land, vegetation.Fasme     (level, random.randint(2, 8)))
  for i in range(4): place(level, land, vegetation.TreePompon(level))
  land.reinit() # XXX in end_land ???
  
  random.shuffle(level.interesting_places)
  for i in range( 4): add_random_item(level, land)
  
  c = npc.Morkul1(0)
  p = soya.Point(houses[0], 6.0, 0.0, 2.0)
  place_character(level, land, c, p)
  c.queue.controler = npc.MorkulAI(c, p)
  
  c = npc.Morkul2(0)
  p = soya.Point(houses[1], 6.0, 0.0, 2.0)
  place_character(level, land, c, p)
  c.queue.controler = npc.MorkulAI(c, p)
    
  c = npc.Morkul3(0)
  p = soya.Point(houses[2], 6.0, 0.0, 2.0)
  place_character(level, land, c, p)
  c.queue.controler = npc.MorkulAI(c, p)
    
  c = npc.Morkar(0)
  p = soya.Point(morkul_temple, 0.0, 0.0, -4.0)
  place_character(level, land, c, p)
  c.queue.controler = npc.MorkulAI(c, p)
  level.axe = c
  
  end_level(level)
  
  return level

