# Balazar
# Copyright (C) 2003 Jean-Baptiste LAMY
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import random, math

import soya
from soya import Point, Vector

import balazar.base      as base
import balazar.character as character
import balazar.sound     as sound

UP   = Vector(None, 0.0,  1.0, 0.0)
DOWN = Vector(None, 0.0, -1.0, 0.0)

class Vegetal(base.Entity):
  pass


MUSHROOM_SHAPES = (soya.Shape.get("champignon@pos1"), soya.Shape.get("champignon@pos2"), soya.Shape.get("champignon@pos3"))
#MUSHROOM_SHAPES = (soya.Shape.get("champignon_0"), soya.Shape.get("champignon_1"), soya.Shape.get("champignon_2"))
MUSHROOM_CAL3D_SHAPE = soya.Cal3dShape.get("champignon")

class Mushroom(soya.World, Vegetal, base.Strikeable, base.PhotoScoreable):
  radius = 1.0
  
  def __init__(self, parent = None, state = -0.8, grow = 1.0, size = 1.0):
    soya.World.__init__(self, parent)
    self.volume    = soya.Volume(self, MUSHROOM_SHAPES[int(state)])
    self.volume.name = "mushroom.volume"
    self.state     = state
    self.grow      = grow
    self.animating = 0
    self.scale(0.2176291357901485, 0.2176291357901485, 0.2176291357901485)
    self.sub_mushrooms = []
    self.parent_mushroom = None
    self.center = Point(self, 0.0, 1.2, 0.0)
    
    self.scale(size, size, size)
    
  def get_photo_score(self, photo, camera):
    score = list(base.SHAPES_SCORES.get(self.volume.shape.filename))
    if not score: return base.DEFAULT_SCORE
    score[2] = self.scale_x
    return score
  
  def kesako(self): return _("__kesako__champignon")
  
  def hurt(self, caster, damage):
    if isinstance(self.parent, base.EjectedBody): return 1
    if isinstance(self.parent, base.HurtedBody ): return 0
    
    if self.state < 3.0:
      sound.play("morfle1.wav", self, gain = 4.0)
      if damage > -5.0:
        self.kill(caster)
        return 1
      
      Spore(self.parent, 10).move(self)
      
      for mushroom in self.sub_mushrooms: mushroom.hurt(caster, damage)
      
      base.HurtedBody(self.parent, self, caster >> self)
      return 0
      
  def kill(self, caster):
    if isinstance(self.parent, base.EjectedBody): return
    
    if self.state < 3.0:
      
      Spore(self.parent, 16).move(self)
      
      if self.parent_mushroom:
        self.parent_mushroom.sub_mushrooms.remove(self)
        if not self.parent_mushroom.sub_mushrooms: self.parent_mushroom.grow = 1.0
        
      for mushroom in self.sub_mushrooms[:]: mushroom.kill(caster)
      
      base.EjectedBody(self.parent, self, caster >> self)
      
  def begin_round(self):
    soya.World.begin_round(self)
    
    if self.animating < 0: # Disappear
      self.animating -= 1
      self.scale(0.9, 0.9, 0.9)
      if self.animating < -15: self.parent.remove(self)
    elif self.state < 0.0: # Appear
      self.state += 0.05
      self.scale(1.1, 1.1, 1.1)
      
  def big_round(self):
    if self.state < 0.0: return
    
    if self.animating > 0: # Playing animation
      self.animating -= 1
      if self.animating == 0:
        self.remove(self.perso)
        self.volume.set_shape(MUSHROOM_SHAPES[int(self.state)])
        self.volume.visible = self.volume.solid = 1
      return
      
    old_state = self.state
    self.state += self.grow
    
    if old_state < 1.0 <= self.state:
      self.animating = 2
      self.perso = soya.Cal3dVolume(self, MUSHROOM_CAL3D_SHAPE)
      self.perso.animate_execute_action("pousse1", 0.0, 0.0)
      self.volume.visible = self.volume.solid = 0
      
    elif old_state < 2.0 <= self.state:
      self.animating = 2
      self.perso = soya.Cal3dVolume(self, MUSHROOM_CAL3D_SHAPE)
      self.perso.animate_execute_action("pousse2", 0.0, 0.0)
      self.volume.visible = self.volume.solid = 0
      
    elif old_state < 3.0 <= self.state:
      sound.play("morfle1.wav", self, gain = 4.0)
      parent = self.parent
      spore = Spore(parent)
      spore.move(self)
      
      self.animating = -1
      
      if self.parent_mushroom:
        self.parent_mushroom.sub_mushrooms.remove(self)
        if not self.parent_mushroom.sub_mushrooms: self.parent_mushroom.grow = 1.0
        
      if len(parent.children) < 50:
        self.solid = 0
        context = parent.get_root().RaypickContext(self, 10.0)
        self.solid = 1
        if self.scale_x > 2.0:
          for obj in self.parent:
            if isinstance(obj, GiantMushroom): break # Only ONE giant mushroom per level
          else:
            child = GiantMushroom()
            child.rotate_lateral(parent.random.uniform(0.0, 360.0))
            child.set_xyz(
              self.x + parent.random.uniform(-5.0, 5.0),
              self.y + 10.0,
              self.z + parent.random.uniform(-5.0, 5.0),
              )
            parent.add(child)
            r = context.raypick(child, DOWN, -1.0, 1, 1)
            if r:
              r[0].convert_to(parent)
              child.y = r[0].y - 1.0
            else: parent.remove(child)
            
        else:
          self.solid = 0
          for i in range(parent.random.randint(1, 3)):
            child = Mushroom(None, grow = parent.random.uniform(0.01, 0.1), size = max(0.4, self.scale_x + parent.random.uniform(-0.3, 0.3)))
            child.rotate_lateral(parent.random.uniform(0.0, 360.0))
            child.set_xyz(
              self.x + parent.random.uniform(-5.0, 5.0),
              self.y + 10.0,
              self.z + parent.random.uniform(-5.0, 5.0),
              )
            parent.add(child)
            r = context.raypick(child, DOWN, -1.0, 1, 1)
            if r:
              #if   getattr(r[0].parent, "name", "") == "mushroom.volume":
              if   isinstance(r[0].parent.parent, character.Character):
                parent.remove(child)
                continue
              elif isinstance(r[0].parent.parent, Mushroom):
                r[0].parent.parent.grow = 0.0
                r[0].parent.parent.sub_mushrooms.append(child)
                child.parent_mushroom = r[0].parent.parent
              elif getattr(r[0].parent, "name", "") == "giant_mushroom.volume":
                parent.remove(child)
                continue
              r[0].convert_to(parent)
              child.y = r[0].y
              
              #v = r[1].cross_product(UP)
              #v.convert_to(self)
              #child.rotate_axe(-r[1].angle_to(UP), v.x, v.y, v.z)
            else: parent.remove(child)
          self.solid = 1

      
    
      
      
class GiantMushroom(soya.World, Vegetal):
  radius = 10.0
  
  def __init__(self, parent = None, state = -0.8):
    soya.World.__init__(self, parent)
    if state < 1.0: self.volume = soya.Volume(self, soya.Shape.get("champignon_geant@pos1"))
    else:           self.volume = soya.Volume(self, soya.Shape.get("champignon_geant@pos2"))
    self.volume.name = "giant_mushroom.volume"
    self.state     = state
    self.scale(0.2176291357901485, 0.2176291357901485, 0.2176291357901485)
    self.animating = 0
    
  def display_name(self):
    if   self.state < 1.0: return _("__shape__champignon_geant_0")
    elif self.state < 3.0: return _("__shape__champignon_geant")
    else:                  return _("__shape__vieux_champignon_geant")
    
  def kesako(self):
    if self.state < 3.0: return _("__kesako__champignon_geant")
    else:                return _("__kesako__vieux_champignon_geant")
    
  def get_photo_score(self, photo, camera):
    return 20.0 + 20.0 * self.state, 10.0 * self.state, 1.0 + 2.0 * self.state
  
  def begin_round(self):
    soya.World.begin_round(self)
    
    if self.state < 0.0: # Appear
      self.state += 0.05
      self.scale(1.1, 1.1, 1.1)
    else:
      old_state = self.state
      
      if self.state < 3.0:
        if self.animating > 0: # Playing animation
          self.animating -= 1
          if self.animating == 0:
            self.remove(self.perso)
            self.volume.set_shape(soya.Shape.get("champignon_geant@pos2"))
            self.volume.visible = self.volume.solid = 1
          return
        
        self.state += 0.001
        
        if   old_state < 1.0 <= self.state:
          self.animating = 35
          self.perso = soya.Cal3dVolume(self, soya.Cal3dShape.get("champignon_geant"))
          self.perso.animate_execute_action("pousse1", 0.0, 0.0)
          self.volume.visible = self.volume.solid = 0
          
  def advance_time(self, proportion):
    soya.World.advance_time(self, proportion)
    
    if (not self.animating) and (0.0 <= self.state < 3.0):
      self.scale_x = self.scale_y = self.scale_z = self.scale_x + proportion * 0.002
      
  
class PlantElement(soya.World, base.PhotoScoreable):
  def __init__(self, parent = None, shape = None):
    soya.World.__init__(self, parent, shape)
    
  def get_photo_score(self, photo, camera):
    return base.SHAPES_SCORES.get(self.shape.filename) or base.DEFAULT_SCORE
  


FASME_SHAPE_NAMES = [
  "fasme@var1",
  "fasme@var2",
  ]
class Fasme(soya.World, Vegetal):
  radius = 5.0
  
  def __init__(self, parent = None, size = 8):
    soya.World.__init__(self, parent)
    self.under_ground = 0
    self.appearing    = 0
    self.oscille_angle = self.oscille_angle2 = 0.0
    self.oscille_vx = 1.0
    self.oscille_vy = 0.0
    self.oscille_vz = 0.0
    self.size       = size
    
  def kesako(self): return _("__kesako__fasme")

  def placed(self):
    scene = self.parent.get_root() # Can't use context here !!! since we are adding items...
    
    previous = None
    for i in range(self.size):
      child = PlantElement(self, soya.Shape.get(self.parent.random.choice(FASME_SHAPE_NAMES)))
      child.rotate_lateral(self.parent.random.uniform(0.0, 360.0))
      child.set_xyz(
        self.parent.random.uniform(-5.0, 5.0),
        (i + 1) * 10.0,
        self.parent.random.uniform(-5.0, 5.0),
        )
      child.solid = 0
      for j in range(4):
        r = scene.raypick(child, DOWN, -1.0, 1, 1)
        if r and ((not previous) or (r[0].parent is previous)): break
      else: self.remove(child); continue
      child.solid = 1
      
      if isinstance(r[0].parent, PlantElement):
        self.remove(child)
        r[0].parent.add(child)
        
      child.move(r[0])
      
      child.oscille_vx, child.oscille_vy, child.oscille_vz = child.transform_vector(self.oscille_vx, self.oscille_vy, self.oscille_vz, self)
      previous = child
      
  def appear_element(self, element):
    if element.scale_x < 1.0:
      if not element.visible: element.visible = 1
      element.scale(1.0 / 0.9, 1.0 / 0.9, 1.0 / 0.9)
      return 1
    
    for child in element.children:
      if self.appear_element(child): return 1
    return 0
  
  def disappear_element(self, element):
    for child in element.children:
      if self.disappear_element(child): return 1
      
    if element.visible:
      element.scale(0.9, 0.9, 0.9)
      if element.scale_x < 0.1: element.visible = 0
      return 1
    
    return 0

  def oscille_elements(self, elements, angle):
    for child in elements.children:
      if child.scale_x >= 1.0:
        child.rotate_axe_xyz(angle, elements.oscille_vx, elements.oscille_vy, elements.oscille_vz)
        if child.children: self.oscille_elements(child, angle)
        
  #def oscille_elements_fix(self, elements, angle):
  #  for child in elements.children:
  #    if child.scale_x >= 1.0:
  #      child._matrix = child.base_pos
  #      child.rotate_axe(angle, elements.oscille_vx, elements.oscille_vy, elements.oscille_vz)
  #      if child.children: self.oscille_elements_fix(child, angle)
        
  def begin_round(self):
    if   self.appearing ==  1:
      for child in self.children:
        if self.appear_element(child): break
      else:
        self.appearing = self.under_ground = 0
        
    elif self.appearing == -1:
      for child in self.children:
        if self.disappear_element(child): break
      else:
        self.under_ground = 1
        self.visible = self.solid = self.appearing = 0
        
    if self.visible:
      self.oscille_angle  += 0.01 
      self.oscille_angle2 += 0.1 * math.cos(self.oscille_angle)
      
  def advance_time(self, proportion):
    if self.visible: self.oscille_elements(self, proportion * 0.1 * math.cos(self.oscille_angle))
    
  #def end_round(self):
  #  if self.visible: self.oscille_elements_fix(self, self.oscille_angle2)
    
  def big_round(self):
    if self.appearing == 0:
      if self.under_ground:
        if self.parent.random.random() < 0.03:
          self.visible = self.solid = 1
          self.appearing = 1
        else:
          for character in self.parent.characters.values():
            if character.distance_to(self) < 20.0:
              self.visible = self.solid = 1
              self.appearing = 1
              break
            
      else:
        if self.parent.random.random() < 0.003: self.appearing = -1
        else:
          for character in self.parent.characters.values():
            if character.distance_to(self) < 50.0: break
          else: self.appearing = -1
          
    
    



class TreePompon(soya.Volume, Vegetal, base.Strikeable, base.PhotoScoreable, base.Terraformer):
  radius           = 5.5
  terraform_radius = 3.0
  
  def __init__(self, parent = None):
    soya.Volume.__init__(self, parent, soya.Shape.get("scn-tree5"))

  def hurt(self, caster, damage):
    if isinstance(self.parent, base.HurtedBody ): return 0
    
    base.HurtedBody(self.parent, self, caster >> self)
    return 0
  
  
  
  
import os, os.path, cPickle as pickle
#SPORE_MATERIAL = pickle.load(open(os.path.join(soya.DATADIR, "fx.data"), "rb"))
SPORE_MATERIAL = soya.PARTICLE_DEFAULT_MATERIAL

class Spore(soya.Particles):
  def __init__(self, parent = None, nb_particles = 15):
    #soya.Particles.__init__(self, parent, particle._default() or SPORE_MATERIAL, nb_particles)
    soya.Particles.__init__(self, parent, SPORE_MATERIAL, nb_particles)
    self.auto_generate_particle = 1
    self.removable = 1
    self.set_colors((0.5, 0.95, 0.05, 0.5), (0.5, 0.95, 0.05, 1.0), (0.0, 0.7, 0.0, 0.2))
    self.set_sizes((2.0, 2.0))
    self.life = 25
    
  def generate(self, index):
    angle = random.random() * 3.1417
    sx = math.cos(angle)
    sy = math.sqrt(random.random() * 3.0)
    sz = math.sin(angle)
    l = (0.06 * (1.5 + random.random())) / math.sqrt(sx * sx + sy * sy + sz * sz)
    self.set_particle(index, 0.5 + random.random(), sx * l, sy * l, sz * l, 0.0, 0.01, 0.0)
    
  def begin_round(self):
    soya.Particles.begin_round(self)
    if   self.life >  0: self.life -= 1
    elif self.life == 0:
      self.auto_generate_particle = 0
      self.life = -1


